"""Core context processors."""

from django.conf import settings
from django.contrib.sites.shortcuts import get_current_site

from geokey import version


def project_settings(request):
    GOOGLE_ANALYTICS = None

    if hasattr(settings, 'GOOGLE_ANALYTICS'):
        GOOGLE_ANALYTICS = settings.GOOGLE_ANALYTICS

    return {
        'DEBUG': settings.DEBUG,
        'PLATFORM_NAME': get_current_site(request).name,
        'GEOKEY_VERSION': version.get_version(),
        'GOOGLE_ANALYTICS': GOOGLE_ANALYTICS
    }



"""Serializers for contributions."""

import requests
import tempfile

from django.core import files
from django.core.exceptions import PermissionDenied, ValidationError

from easy_thumbnails.files import get_thumbnailer
from easy_thumbnails.exceptions import InvalidImageFormatError

from rest_framework import serializers
from rest_framework_gis import serializers as geoserializers

from rest_framework.serializers import BaseSerializer

from geokey.categories.serializers import CategorySerializer
from geokey.categories.models import Category
from geokey.users.serializers import UserSerializer

from .models import (
    Observation,
    Location,
    Comment,
    MediaFile,
    ImageFile,
    VideoFile,
    AudioFile
)


class LocationSerializer(geoserializers.GeoFeatureModelSerializer):
    """
    Serialiser for geokey.contribtions.models.Location
    """
    class Meta:
        model = Location
        geo_field = 'geometry'
        fields = ('id', 'name', 'description', 'status', 'created_at')
        write_only_fields = ('status',)


class LocationContributionSerializer(serializers.ModelSerializer):
    """
    Serialiser for `Location`; to be used within `ContributionSerializer`.
    """
    class Meta:
        model = Location
        fields = ('id', 'name', 'description', 'status', 'created_at',
                  'geometry', 'private', 'private_for_project')
        write_only_fields = ('status', 'private', 'private_for_project')

    def create(self, validated_data):
        """
        Creates a new contribution from `validated_data`

        Parameter
        ---------
        validated_data : dict
            Input data after validation

        Returns
        -------
        Location
        """
        validated_data['creator'] = self.context.get('user')
        return super(
            LocationContributionSerializer,
            self
        ).create(validated_data)


class ContributionSerializer(BaseSerializer):
    """
    Serialiser for geokey.contribtions.models.Observations. This is a custom
    serialiser, not a standard ModelSerializer
    """
    @classmethod
    def many_init(cls, *args, **kwargs):
        """
        Is called when many=True property is set when instantiating the
        serialiser.
        """
        kwargs['context']['many'] = True
        return super(ContributionSerializer, cls).many_init(*args, **kwargs)

    def validate_category(self, project, category_id):
        """
        Validates if the category can be used with the project

        Parameters
        ----------
        project : geokey.projects.models.Project
            Project that the category is used for
        category_id : int
            identifies the category in the database

        Returns
        -------
        geokey.categories.models.Category
            The valid category
        """
        errors = []
        category = None
        try:
            category = project.categories.get(pk=category_id)
            if category.status == 'inactive':
                errors.append('The category can not be used because it is '
                              'inactive.')
            else:
                self._validated_data['meta']['category'] = category
        except Category.DoesNotExist:
            errors.append('The category can not be used with the project '
                          'or does not exist.')

        if errors:
            self._errors['category'] = errors

        return category

    def replace_null(self, properties):
        """
        Replaces all empty str or unicode values with None and returns the
        properies dict

        Parameter
        ---------
        properties : dict
            Contribution properties

        Returns
        -------
        dict
            Contribution properties with replaced null values

        """
        for key, value in properties.iteritems():
            if isinstance(value, (str, unicode)) and len(value) == 0:
                properties[key] = None

        return properties

    def validate_properties(self, properties, category=None, status=None):
        """
        Validates the properties and adds error messages to self._errors

        Parameter
        ---------
        properties : dict
            Contribution properties
        category : geokey.categories.models.Category
            Category the properties are validated against
        status : str
            Status for the contribution
        """
        errors = []

        if self.instance:
            status = status or self.instance.status
            if self.instance.properties:
                update = self.instance.properties.copy()
                update.update(properties)
                properties = update
        else:
            status = status or category.default_status

        properties = self.replace_null(properties)
        try:
            if status == 'draft':
                Observation.validate_partial(category, properties)
            else:
                Observation.validate_full(category, properties)
        except ValidationError, e:
            errors.append(e)

        self._validated_data['properties'] = properties
        self._validated_data['meta']['status'] = status

        if errors:
            self._errors['properties'] = errors

    def validate_location(self, project, location_id):
        """
        Validates if the location can be used with the project

        Parameters
        ----------
        project : geokey.projects.models.Project
            Project that the category is used for
        location_id : int
            identifies the location in the database
        """
        errors = []
        self.location = None

        try:
            if location_id is not None:
                self.location = Location.objects.get_single(
                    self.context.get('user'),
                    project.id,
                    location_id
                )
        except PermissionDenied, error:
            errors.append(error)
        except Location.DoesNotExist, error:
            errors.append(error)

        if errors:
            self._errors['location'] = errors

    def is_valid(self, raise_exception=False):
        """
        Checks if the contribution that is deserialised is valid. Validates
        location, category and properties.

        Parameter
        ---------
        raise_exception : Boolean
            indicates if an exeption should be raised if the data is invalid.
            If set to false, this method will return False if the data is
            invalid.

        Returns
        -------
        Boolean
            indicating if data is valid

        Raises
        ------
        ValidationError
            If data is invalid. Exception is raised when raise_exception is set
            tp True.
        """
        self._errors = {}
        self._validated_data = self.initial_data

        project = self.context.get('project')
        meta = self.initial_data.get('meta')

        if meta is None:
            self._validated_data['meta'] = dict()

        # Validate location
        location_id = None
        if self.initial_data.get('location') is not None:
            location_id = self.initial_data.get('location').get('id')
        self.validate_location(project, location_id)

        # Validate category
        category = None
        if self.instance is None and meta is not None:
            category = self.validate_category(project, meta.get('category'))
        else:
            category = self.instance.category
        self._validated_data['meta']['category'] = category

        # Validatie properties
        properties = self.initial_data.get('properties') or {}
        status = None
        if meta is not None:
            status = meta.get('status', None)
        if properties is not None and category is not None:
            self.validate_properties(
                properties,
                category=category,
                status=status
            )

        # raise the exception
        if self._errors and raise_exception:
            raise ValidationError(self._errors)

        return not bool(self._errors)

    def create(self, validated_data):
        """
        Creates a new observation and returns the instance.

        Parameter
        ---------
        validated_data : dict
            the data dict after validation

        Returns
        -------
        geokey.contributions.models.Observation
            The instance created
        """
        project = self.context.get('project')
        meta = validated_data.pop('meta')

        location_serializer = LocationContributionSerializer(
            self.location,
            data=validated_data.pop('location', None),
            context=self.context
        )
        if location_serializer.is_valid():
            location_serializer.save()

        self.instance = Observation.create(
            properties=validated_data.get('properties'),
            creator=self.context.get('user'),
            location=location_serializer.instance,
            project=project,
            category=meta.get('category'),
            status=meta.pop('status', None)
        )

        return self.instance

    def update(self, instance, validated_data):
        """
        Updates an existing observation and returns the instance.

        Parameter
        ---------
        instance : geokey.contributions.models.Observation
            the instance to be updated
        validated_data : dict
            the data dict after validation

        Returns
        -------
        geokey.contributions.models.Observation
            The instance update
        """
        meta = validated_data.get('meta')
        status = None
        if meta is not None:
            status = meta.get('status', None)

        location_serializer = LocationContributionSerializer(
            instance.location,
            data=validated_data.pop('location', {}),
            context=self.context,
            partial=True
        )
        if location_serializer.is_valid():
            location_serializer.save()

        return instance.update(
            properties=validated_data.get('properties'),
            updator=self.context.get('user'),
            status=status
        )

    def get_display_field(self, obj):
        """
        Returns a native representation of the display_field property.

        Parameter
        ---------
        obj : geokey.contributions.models.Observation
            The instance that is serialised

        Returns
        -------
        dict
            serialised display_field; e.g.:
            {
                'key': 'field_key',
                'value': 'The value of the field'
            }
        """
        if obj.display_field is not None:
            display_field = obj.display_field.split(':', 1)
            value = display_field[1] if display_field[1] != 'None' else None
            return {
                'key': display_field[0],
                'value': value
            }
        else:
            return None

    def get_expiry_field(self, obj):
        """
        Returns a native representation of the expiry_field property.

        Parameter
        ---------
        obj : geokey.contributions.models.Observation
            The instance that is serialised

        Returns
        -------
        dict
            serialised expiry_field; e.g.: '2016-06-02 10:41:52.929103+00:00'
        """
        return str(obj.expiry_field) if obj.expiry_field else None

    def to_representation(self, obj):
        """
        Returns the native representation of a contribution

        Parameter
        ---------
        obj : geokey.contributions.models.Observation
            The instance that is serialised

        Returns
        -------
        dict
            Native represenation of the Contribution
        """
        location = obj.location

        isowner = False
        if not self.context.get('user').is_anonymous():
            isowner = obj.creator == self.context.get('user')

        updator = None
        if obj.updator is not None:
            updator = {
                'id': obj.updator.id,
                'display_name': obj.updator.display_name
            }

        feature = {
            'id': obj.id,
            'properties': obj.properties,
            'display_field': self.get_display_field(obj),
            'expiry_field': self.get_expiry_field(obj),
            'meta': {
                'status': obj.status,
                'creator': {
                    'id': obj.creator.id,
                    'display_name': obj.creator.display_name
                },
                'updator': updator,
                'created_at': str(obj.created_at),
                'updated_at': str(obj.updated_at),
                'version': obj.version,
                'isowner': isowner,
                'num_media': obj.num_media,
                'num_comments': obj.num_comments
            },
            'location': {
                'id': location.id,
                'name': location.name,
                'description': location.description,
                'geometry': location.geometry.geojson
            }
        }

        if self.context.get('many'):
            cat = obj.category
            feature['meta']['category'] = {
                'id': cat.id,
                'name': cat.name,
                'description': cat.description,
                'symbol': cat.symbol.url if cat.symbol else None,
                'colour': cat.colour
            }

        else:
            category_serializer = CategorySerializer(
                obj.category, context=self.context)
            feature['meta']['category'] = category_serializer.data

            comment_serializer = CommentSerializer(
                obj.comments.filter(respondsto=None),
                many=True,
                context=self.context
            )
            feature['comments'] = comment_serializer.data

            review_serializer = CommentSerializer(
                obj.comments.filter(review_status='open'),
                many=True,
                context=self.context
            )
            feature['review_comments'] = review_serializer.data

            file_serializer = FileSerializer(
                obj.files_attached.all(),
                many=True,
                context=self.context
            )
            feature['media'] = file_serializer.data

        return feature


class CommentSerializer(serializers.ModelSerializer):
    """
    Serialiser for geokey.contributions.models.Comment
    """
    creator = UserSerializer(fields=('id', 'display_name'), read_only=True)
    isowner = serializers.SerializerMethodField()

    class Meta:
        model = Comment
        fields = ('id', 'respondsto', 'created_at', 'text', 'isowner',
                  'creator', 'review_status')
        read_only = ('id', 'respondsto', 'created_at')

    def to_representation(self, obj):
        """
        Returns native represenation of the Comment. Adds responses to comment

        Parameter
        ---------
        obj : geokey.contributions.models.Comment
            The instance that is serialised

        Returns
        -------
        dict
            Native represenation of the Comment

        """
        native = super(CommentSerializer, self).to_representation(obj)
        native['responses'] = CommentSerializer(
            obj.responses.all(),
            many=True,
            context=self.context
        ).data

        return native

    def get_isowner(self, comment):
        """
        Returns True if the user serialising the Comment has created the
        comment

        Parameter
        ---------
        comment : geokey.contributions.models.Comment
            The instance that is serialised

        Returns
        -------
        Boolean
            indicating of user is creator of comment
        """
        if not self.context.get('user').is_anonymous():
            return comment.creator == self.context.get('user')
        else:
            return False


class FileSerializer(serializers.ModelSerializer):
    """
    Serialiser for geokey.contributions.models.MediaFile instances
    """
    creator = UserSerializer(fields=('id', 'display_name'))
    isowner = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()
    file_type = serializers.SerializerMethodField()
    thumbnail_url = serializers.SerializerMethodField()

    class Meta:
        model = MediaFile
        fields = (
            'id', 'name', 'description', 'created_at', 'creator', 'isowner',
            'url', 'thumbnail_url', 'file_type'
        )

    def get_file_type(self, obj):
        """
        Returns the type of the MediaFile

        Parameter
        ---------
        obj : geokey.contributions.models.MediaFile
            The instance that is serialised

        Returns
        -------
        str
            The type of the, e.g. 'ImageFile'
        """
        return obj.type_name

    def get_isowner(self, obj):
        """
        Returns `True` if the user provided in the serializer context is the
        creator of this file

        Parameter
        ---------
        obj : geokey.contributions.models.MediaFile
            The instance that is serialised

        Returns
        -------
        Boolean
            indicating if user created the file
        """
        if not self.context.get('user').is_anonymous():
            return obj.creator == self.context.get('user')
        else:
            return False

    def get_url(self, obj):
        """
        Return the url to access this file based on its file type

        Parameter
        ---------
        obj : geokey.contributions.models.MediaFile
            The instance that is serialised

        Returns
        -------
        str
            The URL to embed the file on client side
        """
        if isinstance(obj, ImageFile):
            return obj.image.url
        elif isinstance(obj, VideoFile):
            return obj.youtube_link
        elif isinstance(obj, AudioFile):
            return obj.audio.url

    def _get_thumb(self, image, size=(300, 300)):
        """
        Returns the thumbnail of the media file base on the size privoded

        Parameter
        ---------
        image : Image
            The image to be thumbnailed
        size : tuple
            width and height of the thumbnail, defaults to 300 by 300

        Returns
        -------
        Image
            The thumbnail

        """
        thumbnailer = get_thumbnailer(image)
        thumb = thumbnailer.get_thumbnail({
            'crop': True,
            'size': size
        })
        return thumb

    def get_thumbnail_url(self, obj):
        """
        Creates and returns a thumbnail for the MediaFile object

        Parameter
        ---------
        obj : geokey.contributions.models.MediaFile
            The instance that is serialised

        Returns
        -------
        str
            The url to embed thumbnails on client side
        """
        if isinstance(obj, ImageFile):
            # Some of the imported image files in the original community maps
            # seem to be broken. The error thrown when the image can not be
            # read is caught here.
            try:
                return self._get_thumb(obj.image).url
            except (
                IOError,
                InvalidImageFormatError
            ):
                return ''

        elif isinstance(obj, VideoFile):
            if obj.thumbnail:
                # thumbnail has been downloaded, return the link
                return self._get_thumb(obj.thumbnail).url

            request = requests.get(
                'http://img.youtube.com/vi/%s/0.jpg' % obj.youtube_id,
                stream=True
            )

            if request.status_code != requests.codes.ok:
                # Image not found, return placeholder thumbnail
                return '/static/img/play.png'

            lf = tempfile.NamedTemporaryFile()
            # Read the streamed image in sections
            for block in request.iter_content(1024 * 8):

                # If no more file then stop
                if not block:
                    break

                # Write image block to temporary file
                lf.write(block)

            file_name = obj.youtube_id + '.jpg'
            obj.thumbnail.save(file_name, files.File(lf))

            from PIL import Image

            w, h = Image.open(obj.thumbnail).size

            thumb = self._get_thumb(obj.thumbnail, size=(h, h))
            obj.thumbnail.save(file_name, thumb)

            return self._get_thumb(obj.thumbnail).url

        elif isinstance(obj, AudioFile):
            return '/static/img/play.png'



"""Core base."""

from model_utils import Choices


STATUS_ACTION = Choices('created', 'updated', 'deleted')
LOG_MODELS = {
    'Project': [
        'name',
        'status',
        'isprivate',
        'islocked',
        'everyone_contributes',
        'geographic_extent',
    ],
    'Admins': [],
    'UserGroup': [
        'name',
        'can_contribute',
        'can_moderate',
    ],
    'Category': [
        'name',
        'status',
        'default_status',
    ],
    'TextField': [
        'name',
        'status',
        'required',
    ],
    'NumericField': [
        'name',
        'status',
        'required',
    ],
    'DateTimeField': [
        'name',
        'status',
        'required',
    ],
    'DateField': [
        'name',
        'status',
        'required',
    ],
    'TimeField': [
        'name',
        'status',
        'required',
    ],
    'LookupField': [
        'name',
        'status',
        'required',
    ],
    'MultipleLookupField': [
        'name',
        'status',
        'required',
    ],
    'Location': [
        'name',
        'geometry',
    ],
    'Observation': [
        'status',
        'properties',
    ],
    'Comment': [
        'status',
    ],
    'ImageFile': [
        'status',
    ],
    'VideoFile': [
        'status',
    ],
    'AudioFile': [
        'status',
    ],
    'Subset': [
        'name',
    ],
}
LOG_M2M_RELATIONS = ['UserGroup_users']



"""Tests for mixins of extensions."""

from django.test import TestCase
from django.views.generic import View
from django.views.generic.base import TemplateResponseMixin

from rest_framework.test import APIRequestFactory

from geokey.users.tests.model_factories import UserFactory
from geokey.extensions.mixins import SuperuserMixin


class ExampleView(SuperuserMixin, TemplateResponseMixin, View):
    """Set up example view."""

    template_name = 'base.html'

    def get(self, request):
        """Set up GET request."""
        return self.render_to_response({
            'country': 'United Kingdom'
        })


class SuperuserMixinTest(TestCase):
    """Test superuser mixin."""

    def setUp(self):
        """Set up test."""
        self.view = ExampleView.as_view()
        self.request = APIRequestFactory().get('http://example.com')

    def test_with_user(self):
        """Test with user."""
        self.request.user = UserFactory.create(**{'is_superuser': False})
        response = self.view(self.request).render()

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'This extension is for superusers only.'
        )

    def test_with_superuser(self):
        """Test with superuser."""
        self.request.user = UserFactory.create(**{'is_superuser': True})
        response = self.view(self.request).render()

        self.assertEqual(response.status_code, 200)
        self.assertNotContains(
            response,
            'This extension is for superusers only.'
        )



"""Tests for logger: model AudioFile."""

from django.test import TestCase

from geokey.core.models import LoggerHistory
from geokey.users.tests.model_factories import UserFactory
from geokey.projects.tests.model_factories import ProjectFactory
from geokey.categories.tests.model_factories import CategoryFactory
from geokey.contributions.tests.model_factories import (
    LocationFactory,
    ObservationFactory,
)
from geokey.contributions.tests.media.model_factories import AudioFileFactory


class LogAudioFileTest(TestCase):
    """Test model AudioFile."""

    def setUp(self):
        """Set up test."""
        self.user = UserFactory.create()
        self.project = ProjectFactory.create(**{
            'creator': self.user})
        self.category = CategoryFactory.create(**{
            'creator': self.user,
            'project': self.project})
        self.location = LocationFactory.create(**{
            'creator': self.user})
        self.observation = ObservationFactory.create(**{
            'creator': self.user,
            'location': self.location,
            'project': self.project,
            'category': self.category})
        self.audiofile = AudioFileFactory.create(**{
            'creator': self.user,
            'contribution': self.observation})

    def test_log_create(self):
        """Test when audio file is created gets created."""
        log_count_init = LoggerHistory.objects.count()
        audiofile = AudioFileFactory.create(**{
            'creator': self.user,
            'contribution': self.observation})

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, None)
        self.assertEqual(log.mediafile, {
            'id': str(audiofile.id),
            'name': audiofile.name,
            'type': 'AudioFile'})
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'created',
            'class': 'MediaFile'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_delete(self):
        """Test when audio file gets deleted."""
        mediafile_id = self.audiofile.id
        mediafile_name = self.audiofile.name
        log_count_init = LoggerHistory.objects.count()
        self.audiofile.delete()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, None)
        self.assertEqual(log.mediafile, {
            'id': str(mediafile_id),
            'name': mediafile_name,
            'type': 'AudioFile'})
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'deleted',
            'class': 'MediaFile',
            'field': 'status',
            'value': 'deleted'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)



"""Tests for models of contributions (media files)."""

from django.test import TestCase

from geokey.contributions.models import (
    ImageFile, VideoFile, AudioFile,
    post_save_media_file_count_update
)
from geokey.contributions.tests.model_factories import ObservationFactory
from geokey.users.tests.model_factories import UserFactory

from .model_factories import get_image


class TestImageFilePostSave(TestCase):
    def test_post_save_image_file_count_update(self):
        observation = ObservationFactory()
        image_file = ImageFile.objects.create(
            name='Test name',
            description='Test Description',
            contribution=observation,
            creator=UserFactory.create(),
            image=get_image()
        )
        ImageFile.objects.create(
            status='deleted',
            name='Test name',
            description='Test Description',
            contribution=observation,
            creator=UserFactory.create(),
            image=get_image()
        )

        post_save_media_file_count_update(ImageFile, instance=image_file)
        self.assertEqual(image_file.contribution.num_media, 1)
        self.assertEqual(image_file.contribution.num_comments, 0)


class ImageFileTest(TestCase):
    def test_get_type_name(self):
        image_file = ImageFile.objects.create(
            name='Test name',
            description='Test Description',
            contribution=ObservationFactory.create(),
            creator=UserFactory.create(),
            image=get_image()
        )
        self.assertEqual(image_file.type_name, 'ImageFile')

    def test_delete_file(self):
        image_file = ImageFile.objects.create(
            name='Test name',
            description='Test Description',
            contribution=ObservationFactory.create(),
            creator=UserFactory.create(),
            image=get_image()
        )
        image_file.delete()
        self.assertEquals(image_file.status, 'deleted')


class TestVideoFilePostSave(TestCase):
    def test_post_save_video_file_count_update(self):
        observation = ObservationFactory()
        video_file = VideoFile.objects.create(
            name='Test name',
            description='Test Description',
            contribution=observation,
            creator=UserFactory.create(),
            video=get_image(),
            youtube_link='http://example.com/1122323',
            swf_link='http://example.com/1122323.swf'
        )
        VideoFile.objects.create(
            status='deleted',
            name='Test name',
            description='Test Description',
            contribution=observation,
            creator=UserFactory.create(),
            video=get_image(),
            youtube_link='http://example.com/1122323',
            swf_link='http://example.com/1122323.swf'
        )

        post_save_media_file_count_update(VideoFile, instance=video_file)
        self.assertEqual(video_file.contribution.num_media, 1)
        self.assertEqual(video_file.contribution.num_comments, 0)


class VideoFileTest(TestCase):
    def test_get_type_name(self):
        video_file = VideoFile.objects.create(
            name='Test name',
            description='Test Description',
            contribution=ObservationFactory.create(),
            creator=UserFactory.create(),
            video=get_image(),
            youtube_link='http://example.com/1122323',
            swf_link='http://example.com/1122323.swf'
        )
        self.assertEqual(video_file.type_name, 'VideoFile')

    def test_delete_file(self):
        video_file = VideoFile.objects.create(
            name='Test name',
            description='Test Description',
            contribution=ObservationFactory.create(),
            creator=UserFactory.create(),
            video=get_image(),
            youtube_link='http://example.com/1122323',
            swf_link='http://example.com/1122323.swf'
        )
        video_file.delete()
        self.assertEquals(video_file.status, 'deleted')


class TestAudioFilePostSave(TestCase):
    def test_post_save_audio_file_count_update(self):
        observation = ObservationFactory()
        audio_file = AudioFile.objects.create(
            name='Test name',
            description='Test Description',
            contribution=observation,
            creator=UserFactory.create(),
            audio=get_image()
        )
        AudioFile.objects.create(
            status='deleted',
            name='Test name',
            description='Test Description',
            contribution=observation,
            creator=UserFactory.create(),
            audio=get_image()
        )

        post_save_media_file_count_update(AudioFile, instance=audio_file)
        self.assertEqual(audio_file.contribution.num_media, 1)
        self.assertEqual(audio_file.contribution.num_comments, 0)


class AudioFileTest(TestCase):
    def test_get_type_name(self):
        audio_file = AudioFile.objects.create(
            name='Test name',
            description='Test Description',
            contribution=ObservationFactory.create(),
            creator=UserFactory.create(),
            audio=get_image()
        )
        self.assertEqual(audio_file.type_name, 'AudioFile')

    def test_delete_file(self):
        audio_file = AudioFile.objects.create(
            name='Test name',
            description='Test Description',
            contribution=ObservationFactory.create(),
            creator=UserFactory.create(),
            audio=get_image()
        )
        audio_file.delete()
        self.assertEquals(audio_file.status, 'deleted')



"""Tests for logger: model TextField."""

from django.test import TestCase

from geokey.core.models import LoggerHistory
from geokey.users.tests.model_factories import UserFactory
from geokey.projects.tests.model_factories import ProjectFactory
from geokey.categories.tests.model_factories import (
    CategoryFactory,
    TextFieldFactory,
)


class LogTextFieldTest(TestCase):
    """Test model TextField."""

    def setUp(self):
        """Set up test."""
        self.user = UserFactory.create()
        self.project = ProjectFactory.create(**{
            'creator': self.user})
        self.category = CategoryFactory.create(**{
            'creator': self.user,
            'project': self.project})
        self.textfield = TextFieldFactory.create(**{
            'category': self.category})

    def test_log_create(self):
        """Test when text field gets created."""
        log_count_init = LoggerHistory.objects.count()
        field = TextFieldFactory.create(**{
            'category': self.category})

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, {
            'id': str(field.id),
            'name': field.name,
            'type': 'TextField'})
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'created',
            'class': 'Field'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_delete(self):
        """Test when text field gets deleted."""
        field_id = self.textfield.id
        field_name = self.textfield.name
        log_count_init = LoggerHistory.objects.count()
        self.textfield.delete()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, {
            'id': str(field_id),
            'name': field_name,
            'type': 'TextField'})
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'deleted',
            'class': 'Field'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_update_name(self):
        """Test when name changes."""
        log_count_init = LoggerHistory.objects.count()
        self.textfield.name = '%s UPDATED' % self.textfield.name
        self.textfield.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, {
            'id': str(self.textfield.id),
            'name': self.textfield.name,
            'type': 'TextField'})
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Field',
            'field': 'name'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_update_status(self):
        """Test when status changes."""
        log_count_init = LoggerHistory.objects.count()

        self.textfield.status = 'inactive'
        self.textfield.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, {
            'id': str(self.textfield.id),
            'name': self.textfield.name,
            'type': 'TextField'})
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Field',
            'field': 'status',
            'value': self.textfield.status})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

        self.textfield.status = 'active'
        self.textfield.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, {
            'id': str(self.textfield.id),
            'name': self.textfield.name,
            'type': 'TextField'})
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Field',
            'field': 'status',
            'value': self.textfield.status})
        self.assertEqual(log_count, log_count_init + 2)
        self.assertEqual(log.historical, None)

    def test_log_update_required(self):
        """Test when required setting changes."""
        log_count_init = LoggerHistory.objects.count()

        self.textfield.required = True
        self.textfield.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, {
            'id': str(self.textfield.id),
            'name': self.textfield.name,
            'type': 'TextField'})
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Field',
            'field': 'required',
            'value': str(self.textfield.required)})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

        self.textfield.required = False
        self.textfield.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, {
            'id': str(self.textfield.id),
            'name': self.textfield.name,
            'type': 'TextField'})
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Field',
            'field': 'required',
            'value': str(self.textfield.required)})
        self.assertEqual(log_count, log_count_init + 2)
        self.assertEqual(log.historical, None)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2017-02-24 15:49
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('projects', '0008_historicalproject'),
        ('categories', '0017_category_expiry_field'),
    ]

    operations = [
        migrations.CreateModel(
            name='HistoricalCategory',
            fields=[
                ('id', models.IntegerField(auto_created=True, blank=True, db_index=True, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('description', models.TextField(blank=True, null=True)),
                ('created_at', models.DateTimeField(blank=True, editable=False)),
                ('order', models.IntegerField(default=0)),
                ('status', models.CharField(choices=[(b'active', b'active'), (b'inactive', b'inactive'), (b'deleted', b'deleted')], default=b'active', max_length=20)),
                ('default_status', models.CharField(choices=[(b'active', b'active'), (b'pending', b'pending')], default=b'pending', max_length=20)),
                ('colour', models.TextField(default=b'#0033ff')),
                ('symbol', models.TextField(max_length=500, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('creator', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('display_field', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='categories.Field')),
                ('expiry_field', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='categories.Field')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('project', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='projects.Project')),
            ],
            options={
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
                'verbose_name': 'historical category',
            },
        ),
    ]



"""Base for contributions."""

from model_utils import Choices


LOCATION_STATUS = Choices('active', 'review')
OBSERVATION_STATUS = Choices('active', 'draft', 'review', 'pending', 'deleted')
COMMENT_STATUS = Choices('active', 'deleted')
COMMENT_REVIEW = Choices('open', 'resolved')
MEDIA_STATUS = Choices('active', 'deleted')

ACCEPTED_IMAGE_FORMATS = ('png', 'jpeg', 'gif')
ACCEPTED_AUDIO_FORMATS = ('wav', 'wave', 'mp3', 'mpeg', '3gpp', '3gpp2')
ACCEPTED_VIDEO_FORMATS = (
    'mpeg', 'mpeg2', 'mp4', 'quicktime', '3gpp', '3gpp2', 'avi', 'x-flv',
    'x-matroska', 'x-msvideo', 'x-ms-wmv', 'ogg', 'webm'
)



"""Tests for logger: model Comment."""

from django.test import TestCase

from geokey.core.models import LoggerHistory
from geokey.users.tests.model_factories import UserFactory
from geokey.projects.tests.model_factories import ProjectFactory
from geokey.categories.tests.model_factories import CategoryFactory
from geokey.contributions.tests.model_factories import (
    LocationFactory,
    ObservationFactory,
    CommentFactory,
)


class LogCommentTest(TestCase):
    """Test model Comment."""

    def setUp(self):
        """Set up test."""
        self.user = UserFactory.create()
        self.project = ProjectFactory.create(**{
            'creator': self.user})
        self.category = CategoryFactory.create(**{
            'creator': self.user,
            'project': self.project})
        self.location = LocationFactory.create(**{
            'creator': self.user})
        self.observation = ObservationFactory.create(**{
            'creator': self.user,
            'location': self.location,
            'project': self.project,
            'category': self.category})
        self.comment = CommentFactory.create(**{
            'creator': self.user,
            'commentto': self.observation})

    def test_log_create(self):
        """Test when comment gets created."""
        log_count_init = LoggerHistory.objects.count()
        comment = CommentFactory.create(**{
            'creator': self.user,
            'commentto': self.observation})

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, {
            'id': str(comment.id)})
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'created',
            'class': 'Comment'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_create_response(self):
        """Test when response gets created."""
        log_count_init = LoggerHistory.objects.count()
        response = CommentFactory.create(**{
            'creator': self.user,
            'commentto': self.observation,
            'respondsto': self.comment})

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, {
            'id': str(response.id)})
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'created',
            'class': 'Comment',
            'subaction': 'respond',
            'comment_id': str(self.comment.id)})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_delete(self):
        """Test when comment gets deleted."""
        log_count_init = LoggerHistory.objects.count()
        self.comment.delete()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(log.observation, {
            'id': str(self.observation.id)})
        self.assertEqual(log.comment, {
            'id': str(self.comment.id)})
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'deleted',
            'class': 'Comment',
            'field': 'status',
            'value': 'deleted'})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.comment.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.comment.id)

    def test_log_delete_nested(self):
        """Test when comment that has responses gets deleted."""
        response = CommentFactory.create(**{
            'creator': self.user,
            'commentto': self.observation,
            'respondsto': self.comment})
        log_count_init = LoggerHistory.objects.count()
        self.comment.delete()

        log_count = LoggerHistory.objects.count()
        self.assertEqual(log_count, log_count_init + 2)

        logs = LoggerHistory.objects.all().order_by('-pk')[:2]

        # Response gets deleted
        self.assertNotEqual(logs[1].user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(logs[1].project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(logs[1].usergroup, None)
        self.assertEqual(logs[1].category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(logs[1].field, None)
        self.assertEqual(logs[1].location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(logs[1].observation, {
            'id': str(self.observation.id)})
        self.assertEqual(logs[1].comment, {
            'id': str(response.id)})
        self.assertEqual(logs[1].subset, None)
        self.assertEqual(logs[1].action, {
            'id': 'deleted',
            'class': 'Comment',
            'subaction': 'respond',
            'comment_id': str(self.comment.id)})
        self.assertEqual(logs[1].historical, None)

        # Comment gets deleted
        self.assertNotEqual(logs[0].user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(logs[0].project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(logs[0].usergroup, None)
        self.assertEqual(logs[0].category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(logs[0].field, None)
        self.assertEqual(logs[0].location, {
            'id': str(self.location.id),
            'name': self.location.name})
        self.assertEqual(logs[0].observation, {
            'id': str(self.observation.id)})
        self.assertEqual(logs[0].comment, {
            'id': str(self.comment.id)})
        self.assertEqual(logs[0].subset, None)
        self.assertEqual(logs[0].action, {
            'id': 'deleted',
            'class': 'Comment',
            'field': 'status',
            'value': 'deleted'})
        history = self.comment.history.get(pk=logs[0].historical.get('id'))
        self.assertEqual(history.id, self.comment.id)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('projects', '0006_remove_admins_contact'),
    ]

    operations = [
        migrations.RenameField(
            model_name='project',
            old_name='geographic_extend',
            new_name='geographic_extent',
        ),
        migrations.AddField(
            model_name='project',
            name='islocked',
            field=models.BooleanField(default=False),
        ),
    ]



"""Core models."""

from django.db.models.signals import (
    pre_save,
    post_save,
    post_delete,
    m2m_changed,
)
from django.dispatch import receiver
from django.contrib.postgres.fields import HStoreField

from model_utils.models import TimeStampedModel

from geokey.core.signals import get_request

from .base import STATUS_ACTION, LOG_MODELS, LOG_M2M_RELATIONS


class LoggerHistory(TimeStampedModel):
    """Store the event logs."""

    user = HStoreField(null=True, blank=True)
    project = HStoreField(null=True, blank=True)
    usergroup = HStoreField(null=True, blank=True)
    category = HStoreField(null=True, blank=True)
    field = HStoreField(null=True, blank=True)
    location = HStoreField(null=True, blank=True)
    observation = HStoreField(null=True, blank=True)
    comment = HStoreField(null=True, blank=True)
    mediafile = HStoreField(null=True, blank=True)
    subset = HStoreField(null=True, blank=True)
    action = HStoreField(null=True, blank=True)
    historical = HStoreField(null=True, blank=True)


def get_class_name(instance_class):
    """Get the instance class name."""
    if not hasattr(instance_class, '__bases__'):
        instance_class = instance_class.__class__

    bases = [x.__name__ for x in instance_class.__bases__]
    if 'Field' in bases:
        return 'Field'
    if 'MediaFile' in bases:
        return 'MediaFile'
    return instance_class.__name__


def get_history(instance):
    """Get the latest history entry of an instance."""
    try:
        history = {
            'id': str(instance.history.latest('pk').pk),
            'class': instance.history.model.__name__,
        }
    # TODO: Except when history model object does not exist
    except:
        history = None

    return history


def add_extra_info(action, instance):
    """Add the extra instance info to the action."""
    action_class = action.get('class')

    if action_class in ['Admins', 'UserGroup_users']:
        if hasattr(instance, 'user'):
            instance = instance.user
        # Add user details for user groups (including admin groups)
        action['user_id'] = str(instance.id)
        action['user_display_name'] = str(instance)
    elif action_class == 'Comment':
        # There might be that a parent comment does not exist
        try:
            if instance.respondsto:
                # Add parent comment details - handy to track nested comments
                action['subaction'] = 'respond'
                action['comment_id'] = str(instance.respondsto_id)
        except instance.__class__.DoesNotExist:
            pass

    return action


def generate_log(sender, instance, action):
    """Generate a single log (without saving to DB)."""
    log = LoggerHistory(action=action)

    request = get_request()
    if hasattr(request, 'user'):
        log.user = {
            'id': str(request.user.id),
            'display_name': str(request.user),
        }

    fields = {}
    class_name = get_class_name(sender)

    if class_name == 'Project':
        fields['project'] = instance
    elif hasattr(instance, 'project'):
        fields['project'] = instance.project

    if 'UserGroup' in class_name:
        fields['usergroup'] = instance

    if class_name == 'Category':
        fields['category'] = instance
    elif hasattr(instance, 'category'):
        fields['category'] = instance.category

    if class_name == 'Location':
        fields['location'] = instance
    elif hasattr(instance, 'location'):
        fields['location'] = instance.location

    if class_name == 'Field':
        fields['project'] = instance.category.project
        fields['field'] = instance
    elif class_name == 'Observation':
        fields['observation'] = instance
    elif class_name == 'Comment':
        fields['project'] = instance.commentto.project
        fields['category'] = instance.commentto.category
        fields['location'] = instance.commentto.location
        fields['observation'] = instance.commentto
        fields['comment'] = instance
    elif class_name == 'MediaFile':
        fields['project'] = instance.contribution.project
        fields['category'] = instance.contribution.category
        fields['location'] = instance.contribution.location
        fields['observation'] = instance.contribution
        fields['mediafile'] = instance
    elif class_name == 'Subset':
        fields['subset'] = instance

    for field, instance in fields.iteritems():
        value = {'id': str(instance.id)}

        # Not all models have names
        if hasattr(instance, 'name'):
            value['name'] = instance.name

        # Fields for categories and media files should also have type
        if field in ['field', 'mediafile']:
            value['type'] = sender.__name__

        setattr(log, field, value)

    return log


def cross_check_fields(new_instance, old_instance):
    """Check for changed fields between new and old instances."""
    action_id = STATUS_ACTION.updated
    class_name = get_class_name(new_instance)

    changed_fields = []
    usergroup_permission_fields = {}

    for field in LOG_MODELS.get(new_instance.__class__.__name__, {}):
        new_value = new_instance.__dict__.get(field)
        old_value = old_instance.__dict__.get(field)
        if new_value != old_value:
            # Do not log property change when instance (observation) is draft
            if field == 'properties' and new_instance.status == 'draft':
                continue

            # Make specific case for user group permission update
            if field in ['can_contribute', 'can_moderate']:
                usergroup_permission_fields[field] = new_value
                continue

            if field == 'status':
                if old_value == 'draft':
                    # Status changes from "draft" - it's created action
                    action_id = STATUS_ACTION.created
                elif new_value == 'deleted':
                    # Status changes to "deleted" - it's deleted action
                    action_id = STATUS_ACTION.deleted

            changed_field = {
                'id': action_id,
                'class': class_name,
                'field': field,
            }

            if field not in [
                'name',
                'geographic_extent',
                'geometry',
                'properties',
            ]:
                changed_field['value'] = str(new_value)

            changed_fields.append(changed_field)

    if len(usergroup_permission_fields):
        changed_field = {
            'id': action_id,
            'class': class_name,
            'field': 'can_contribute',
            'value': str(True),
        }

        if usergroup_permission_fields.get('can_moderate') is True:
            changed_field['field'] = 'can_moderate'
        elif usergroup_permission_fields.get('can_contribute') is False:
            changed_field['field'] = 'can_view'

        changed_fields.append(changed_field)

    return changed_fields


@receiver(pre_save)
def logs_on_pre_save(sender, instance, **kwargs):
    """Initiate logs when instance get updated."""
    if sender.__name__ in LOG_MODELS:
        logs = []

        try:
            old_instance = sender.objects.get(pk=instance.pk)
            for field in cross_check_fields(instance, old_instance):
                action = add_extra_info(field, instance)
                logs.append(generate_log(sender, instance, action))
        except sender.DoesNotExist:
            pass

        instance._logs = logs


@receiver(post_save)
def log_on_post_save(sender, instance, created, **kwargs):
    """Finalise initiated logs or create a new one when instance is created."""
    if sender.__name__ in LOG_MODELS:
        logs = []

        if created:
            class_name = get_class_name(sender)

            action = add_extra_info({
                'id': STATUS_ACTION.created,
                'class': class_name,
            }, instance)

            if class_name == 'Observation':
                # Do not log new observations when they're still drafts
                if instance.status == 'draft':
                    return
                # We need to know what status observation is when created
                action['field'] = 'status'
                action['value'] = instance.status

            logs.append(generate_log(sender, instance, action))
        elif hasattr(instance, '_logs') and instance._logs is not None:
            logs = instance._logs

        for log in logs:
            log.historical = get_history(instance)
            log.save()


@receiver(post_delete)
def log_on_post_delete(sender, instance, **kwargs):
    """Create a log when instance is deleted."""
    if sender.__name__ in LOG_MODELS:
        action = add_extra_info({
            'id': STATUS_ACTION.deleted,
            'class': get_class_name(sender),
        }, instance)
        log = generate_log(sender, instance, action)
        log.save()


@receiver(m2m_changed)
def log_on_m2m_changed(sender, instance, action, reverse, model, pk_set, **kwargs):
    """Create a log when object is added to or removed from M2M relation."""
    if sender.__name__ in LOG_M2M_RELATIONS and 'post_' in action:
        for pk in pk_set:
            action = add_extra_info({
                'id': STATUS_ACTION.updated,
                'class': sender.__name__,
                'subaction': action.replace('post_', ''),
            }, model.objects.get(pk=pk))
            log = generate_log(sender, instance, action)
            log.save()



"""Tests for core views."""

import json

from django.test import TestCase
from django.http import HttpRequest
from django.contrib.auth.models import AnonymousUser

from geokey.version import get_version
from geokey.core.views import InfoAPIView
from geokey.extensions.base import register, deregister


# ############################################################################
#
# PUBLIC API
#
# ############################################################################

class InfoAPIViewTest(TestCase):
    """Test public API for GeoKey server information."""

    def setUp(self):
        """Set up test."""
        self.view = InfoAPIView.as_view()
        self.request = HttpRequest()
        self.request.method = 'GET'

        register('A', 'A', display_admin=True, superuser=False, version='1.0')
        register('B', 'B', display_admin=True, superuser=False)
        register('S', 'S', display_admin=True, superuser=True, version='0.1.0')

    def tearDown(self):
        """Tear down test."""
        deregister('A')
        deregister('B')
        deregister('S')

    def contains_extension(self, ext_id, installed_extensions):
        """Helper to check if extension is in installed extensions."""
        for extension in installed_extensions:
            if 'name' in extension and extension['name'] == ext_id:
                return True

        return False

    def test_get(self):
        """Test GET."""
        self.request.user = AnonymousUser()
        response = self.view(self.request).render()

        self.assertEqual(response.status_code, 200)

        response = json.loads(response.content)
        self.assertIn('geokey', response)

        geokey = response.get('geokey')
        self.assertEqual(geokey.get('version'), get_version())
        self.assertIn('installed_extensions', geokey)

        installed_extensions = geokey.get('installed_extensions')
        self.assertTrue(self.contains_extension('A', installed_extensions))
        self.assertTrue(self.contains_extension('B', installed_extensions))
        self.assertTrue(self.contains_extension('S', installed_extensions))



"""Template tags for KML."""

from osgeo import ogr

from django import template

from geokey.categories.models import Field, LookupField, MultipleLookupField


register = template.Library()


@register.filter(name='kml_geom')
def kml_geom(place):
    geometry = place.get('location').get('geometry')
    json_geom = ogr.CreateGeometryFromJson(str(geometry))
    kml_geom = json_geom.ExportToKML()
    return kml_geom


@register.filter(name='kml_name')
def kml_name(place):
    name = ''  # Name will always be available, even it's empty

    if place:
        display_field = place.get('display_field')

        if display_field:
            name = display_field.get('value')

    return name


@register.filter(name='kml_desc')
def kml_desc(place):
    properties = place.get('properties')
    media = place.get('media')
    comments = place.get('comments')

    description = '<![CDATA['

    if properties:
        description += '<table>'

        for key in properties:
            name = key
            value = properties[key]

            try:
                field = Field.objects.get(
                    key=key,
                    category_id=place.get('meta').get('category').get('id')
                )
                name = field.name.encode('utf-8')

                if value is not None:
                    if isinstance(field, LookupField):
                        value = field.lookupvalues.get(pk=value).name
                    elif isinstance(field, MultipleLookupField):
                        values = field.lookupvalues.filter(
                            pk__in=value
                        )
                        value = '<br />'.join([v.name for v in values])
            except Field.DoesNotExist:
                pass

            if type(value) in [str, unicode]:
                value = value.encode('utf-8')

            if properties[key] is not None:
                description = '{desc}<tr><td>{name}</td><td>{value}</td></tr>'.format(
                    desc=description,
                    name=name,
                    value=value
                )

        description += '</table>'

    if media:
        description += '<table>'

        for file in media:
            if file['file_type'] == 'ImageFile' or file['file_type'] == 'VideoFile':
                if file['file_type'] == 'VideoFile':
                    file['url'] = file['url'].replace('embed/', 'watch?v=')

                description += '<tr><td><strong>{name}</strong>{desc}<br /><a href="{url}"><img src="{thumbnail_url}" /></a></td></tr>'.format(
                    name=file['name'],
                    desc='<br />%s' % file['description'] if file['description'] else '',
                    url=file['url'],
                    thumbnail_url=file['thumbnail_url']
                )
            elif file['file_type'] == 'AudioFile':
                description += '<tr><td><a href="{url}"><strong>{name}</strong></a>{desc}</td></tr>'.format(
                    name=file['name'],
                    desc='<br />%s' % file['description'] if file['description'] else '',
                    url=file['url']
                )

        description += '</table>'

    if comments:
        def render_comments(comments):
            description = '<table>'

            for comment in comments:
                description += '<tr><td><strong>{name}</strong>{text}'.format(
                    name=comment['creator']['display_name'],
                    text='<br />%s' % comment['text'] if comment['text'] else ''
                )

                if len(comment['responses']) > 0:
                    description += render_comments(comment['responses'])

                description += '</td></tr>'

            description += '</table>'

            return description

        description += render_comments(comments)

    description += ']]>'

    return description


@register.filter(name='kml_style')
def kml_style(place):
    colour = place.get('meta').get('category').get('colour')
    colour = colour.replace('#', '')
    return colour



"""Core serializers."""

from rest_framework import serializers


class FieldSelectorSerializer(serializers.ModelSerializer):
    """
    Field selector serializer.

    Instances accept a `fields` keyword argument to set which fields shall be
    serialized.
    """

    def __init__(self, *args, **kwargs):
        """Initialization."""
        # Don't pass the `fields` argument to the superclass
        fields = kwargs.pop('fields', None)

        # Instantiate the superclass
        super(FieldSelectorSerializer, self).__init__(*args, **kwargs)

        if fields:
            # Drop any fields that are not specified in the `fields` argument
            allowed = set(fields)
            existing = set(self.fields.keys())
            for field in existing - allowed:
                self.fields.pop(field)



"""Core adapters."""

import re

from django.core.urlresolvers import reverse
from django.contrib import messages

from allauth.account.adapter import DefaultAccountAdapter
from allauth.socialaccount.adapter import DefaultSocialAccountAdapter
from allauth.account.models import EmailAddress


class AccountAdapter(DefaultAccountAdapter):
    """Adapter for accounts."""

    username_regex = re.compile(r'^.+$')

    def respond_user_inactive(self, request, user):
        """Resend email confirmation instructions if user is inactive."""
        try:
            email_address = EmailAddress.objects.get(
                user=user,
                email=user.email)
            self.add_message(
                request,
                messages.INFO,
                'account/messages/'
                'email_confirmation_sent.txt',
                {'email': user.email})
            email_address.send_confirmation(request)
        except EmailAddress.DoesNotExist:
            pass

        return super(AccountAdapter, self).respond_user_inactive(request, user)


class SocialAccountAdapter(DefaultSocialAccountAdapter):
    """Adapter for social accounts."""

    def get_connect_redirect_url(self, request, socialaccount):
        """Return URL after successfull connecting a social account."""
        assert request.user.is_authenticated()
        url = reverse('admin:userprofile')
        return url



"""Model factories used for tests of users."""

import datetime
import factory
import string

from factory.fuzzy import FuzzyText

from django.utils import timezone

from oauth2_provider.models import AccessToken

from ..models import User, UserGroup


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = User

    display_name = factory.Sequence(lambda n: "display_name%s" % n)
    email = factory.Sequence(lambda n: "email%s@example.com" % n)
    password = ''
    is_active = True
    is_superuser = False

    last_login = timezone.datetime(2000, 1, 1).replace(tzinfo=timezone.utc)
    date_joined = timezone.datetime(1999, 1, 1).replace(
        tzinfo=timezone.utc)

    @classmethod
    def _prepare(cls, create, **kwargs):
        password = kwargs.pop('password', None)
        user = super(UserFactory, cls)._prepare(create, **kwargs)
        if password:
            user.set_password(password)
            if create:
                user.save()
        return user


# don't move this import to the top of the file! (it would cause an import cycle):
from geokey.applications.tests.model_factories import ApplicationFactory

class AccessTokenFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = AccessToken
        django_get_or_create = ('user', 'application')

    user = factory.SubFactory(UserFactory)
    application = factory.SubFactory(ApplicationFactory)
    token = FuzzyText(length=30, chars=string.ascii_uppercase + string.digits)
    expires = timezone.now() + datetime.timedelta(days=1)
    scope = 'read write'


class UserGroupFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = UserGroup

    name = factory.Sequence(lambda n: 'name_%d' % n)
    description = factory.LazyAttribute(lambda o: '%s description' % o.name)
    project = factory.SubFactory(
        'geokey.projects.tests.model_factories.ProjectFactory')
    can_contribute = True
    can_moderate = False

    @factory.post_generation
    def add_users(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for user in extracted:
                self.users.add(user)



"""Core middleware."""
# https://gist.github.com/barrabinfc/426829

from django import http
from django.db import connection

from .signals import request_accessor

try:
    import settings
    XS_SHARING_ALLOWED_ORIGINS = settings.XS_SHARING_ALLOWED_ORIGINS
    XS_SHARING_ALLOWED_METHODS = settings.XS_SHARING_ALLOWED_METHODS
    XS_SHARING_ALLOWED_HEADERS = settings.XS_SHARING_ALLOWED_HEADERS
except:
    XS_SHARING_ALLOWED_ORIGINS = '*'
    XS_SHARING_ALLOWED_METHODS = ['POST', 'GET', 'OPTIONS', 'PUT', 'PATCH', 'DELETE']
    XS_SHARING_ALLOWED_HEADERS = ['Authorization', 'Content-Type']


class XsSharing(object):
    """
    This middleware allows cross-domain XHR using the HTML5 postMessage API.

    Access-Control-Allow-Origin: http://foo.example
    Access-Control-Allow-Methods: POST, GET, OPTIONS, PUT, DELETE
    """
    def process_request(self, request):
        if 'HTTP_ACCESS_CONTROL_REQUEST_METHOD' in request.META:
            response = http.HttpResponse()
            response['Access-Control-Allow-Origin'] = XS_SHARING_ALLOWED_ORIGINS
            response['Access-Control-Allow-Methods'] = ",".join(XS_SHARING_ALLOWED_METHODS)
            response['Access-Control-Allow-Headers'] = ",".join(XS_SHARING_ALLOWED_HEADERS)
            return response

        return None

    def process_response(self, request, response):
        # Avoid unnecessary work
        if response.has_header('Access-Control-Allow-Origin'):
            return response

        response['Access-Control-Allow-Origin'] = XS_SHARING_ALLOWED_ORIGINS
        response['Access-Control-Allow-Methods'] = ",".join(XS_SHARING_ALLOWED_METHODS)
        response['Access-Control-Allow-Headers'] = ",".join(XS_SHARING_ALLOWED_HEADERS)

        return response


class TerminalLogging(object):
    def process_response(self, request, response):
        from sys import stdout
        if stdout.isatty():
            for query in connection.queries :
                print "\033[1;31m[%s]\033[0m \033[1m%s\033[0m" % (query['time'],
                    " ".join(query['sql'].split()))
        return response


class RequestProvider(object):
    def __init__(self):
        self._request = None
        request_accessor.connect(self)

    def process_view(self, request, view_func, view_args, view_kwargs):
        self._request = request
        return None

    def __call__(self, **kwargs):
        return self._request


def show_debug_toolbar(request):
    """Custom function to determine whether to show the debug toolbar."""
    from django.conf import settings

    # Do not include the debug toolbar on Ajax requests
    if request.is_ajax():
        return False

    return bool(settings.DEBUG and settings.DEBUG_TOOLBAR)



"""Views for observations of categories."""

from django.core.exceptions import PermissionDenied
from django.views.decorators.gzip import gzip_page

from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from geokey.core.decorators import handle_exceptions_for_ajax
from geokey.users.models import User
from geokey.projects.models import Project
from geokey.core.exceptions import InputError

from ..renderers.geojson import GeoJsonRenderer
from ..parsers.geojson import GeoJsonParser

from .base import SingleAllContribution
from ..serializers import ContributionSerializer


class GZipView(object):
    def dispatch(self, request, *args, **kwargs):
        if not hasattr(self, 'META'):
            setattr(self, 'META', {})

        if request.META.get('HTTP_ACCEPT_ENCODING'):
            self.META['HTTP_ACCEPT_ENCODING'] = request.META['HTTP_ACCEPT_ENCODING']

        return super(GZipView, self).dispatch(request, *args, **kwargs)


class GeoJsonView(APIView):
    renderer_classes = (GeoJsonRenderer,)
    parser_classes = (GeoJsonParser,)


class ProjectObservations(GZipView, GeoJsonView):
    """
    Public API endpoint to add new contributions to a project
    /api/projects/:project_id/contributions
    """
    @handle_exceptions_for_ajax
    def post(self, request, project_id):
        """
        Adds a new contribution to a project

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the request
        project_id : int
            identifies the project in the data base

        Returns
        -------
        rest_framework.response.Respone
            Contains the serialised contribution
        """
        user = request.user
        if user.is_anonymous():
            user = User.objects.get(display_name='AnonymousUser')

        data = request.data
        project = Project.objects.as_contributor(request.user, project_id)

        if (not data.get('meta').get('status') == 'draft' and
                project.can_moderate(user)):
            data['meta']['status'] = 'active'

        serializer = ContributionSerializer(
            data=data, context={'user': user, 'project': project}
        )

        if serializer.is_valid(raise_exception=True):
            serializer.save()
            return Response(serializer.data, status=status.HTTP_201_CREATED)

    @gzip_page
    @handle_exceptions_for_ajax
    def get(self, request, project_id):
        """
        Handle GET request.

        Return a list of all contributions of the project accessible to the
        user.

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the request.
        project_id : int
            Identifies the project in the database.

        Returns
        -------
        rest_framework.response.Respone
            Contains the serialized contributions.
        """
        project = Project.objects.get_single(request.user, project_id)
        try:
            contributions = project.get_all_contributions(
                request.user,
                search=request.GET.get('search'),
                subset=request.GET.get('subset'),
                bbox=request.GET.get('bbox')
            ).select_related('location', 'creator', 'updator', 'category') 
        except InputError as e:
            return Response(e, status=status.HTTP_406_NOT_ACCEPTABLE)
        serializer = ContributionSerializer(
            contributions,
            many=True,
            context={
                'user': request.user,
                'project': project,
                'search': request.GET.get('search'),
                'bbox': request.GET.get('bbox')
            }
        )  
        return Response(serializer.data, status=status.HTTP_200_OK)


# ############################################################################
#
# SINGLE CONTRIBUTION
#
# ############################################################################


class SingleContributionAPIView(GeoJsonView):
    """
    Abstract APIView for handling requests to single observations
    """
    def get_and_respond(self, request, observation):
        """
        Returns a single contributions

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the request
        observation : geokey.contributions.models.Observation
            Observation to be returned

        Returns
        -------
        rest_framework.response.Respone
            Contains the serialised observation
        """
        serializer = ContributionSerializer(
            observation,
            context={'user': request.user, 'project': observation.project}
        )
        return Response(serializer.data, status=status.HTTP_200_OK)

    def update_and_respond(self, request, observation):
        """
        Updates and returns a single contributions

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the request
        observation : geokey.contributions.models.Observation
            Observation to be returned

        Returns
        -------
        rest_framework.response.Respone
            Contains the updated serialised observation
        """
        data = request.data
        user = request.user
        if user.is_anonymous():
            user = User.objects.get(display_name='AnonymousUser')

        new_status = None
        if data.get('meta') is not None:
            new_status = data.get('meta').get('status')

        user_can_moderate = observation.project.can_moderate(user)
        user_is_owner = (observation.creator == user)
        under_review = observation.comments.filter(
            review_status='open').exists()

        if (new_status is not None and new_status != observation.status):
            if not (
                (new_status == 'pending' and
                    (user_is_owner or user_can_moderate)) or
                (new_status == 'active' and
                    observation.status == 'draft' and user_is_owner) or
                (new_status == 'active' and
                    observation.status == 'pending' and user_can_moderate)):

                raise PermissionDenied('You are not allowed to update the '
                                       'status of the contribution from "%s" '
                                       'to "%s"' % (
                                           observation.status,
                                           new_status
                                       ))

        elif not (user_is_owner or user_can_moderate):
            raise PermissionDenied('You are not allowed to update the'
                                   'contribution')

        if new_status == 'active' and under_review:
            data['meta']['status'] = 'review'

        if ((new_status == 'active' and observation.status == 'draft') and
                not user_can_moderate):
            default_status = observation.category.default_status
            data['meta']['status'] = default_status

        serializer = ContributionSerializer(
            observation,
            data=data,
            context={'user': user, 'project': observation.project}
        )
        if serializer.is_valid(raise_exception=True):
            serializer.save()
            return Response(serializer.data, status=status.HTTP_200_OK)

    def delete_and_respond(self, request, observation):
        """
        Deletes a single observation

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the request
        observation : geokey.contributions.models.Observation
            Observation to be deleted

        Returns
        -------
        rest_framework.response.Respone
            Empty response indicating successful delete
        """
        if (observation.creator == request.user or
                observation.project.can_moderate(request.user)):
            observation.delete()
            return Response(status=status.HTTP_204_NO_CONTENT)

        raise PermissionDenied('You are not allowed to delete this'
                               'contribution')


class SingleAllContributionAPIView(
        SingleAllContribution, SingleContributionAPIView):
    """
    Public API endpoint for updating a single observation in a project
    /api/projects/:project_id/observations/:observation_id
    """

    @handle_exceptions_for_ajax
    def get(self, request, project_id, observation_id):
        """
        Returns a single contribution

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the request
        project_id : int
            identifies the project in the data base
        observation_id : int
            identifies the observation in the data base

        Returns
        -------
        rest_framework.response.Respone
            Contains the serialised observation
        """
        contribution = self.get_contribution(
            request.user,
            project_id,
            observation_id
        )

        return self.get_and_respond(request, contribution)

    @handle_exceptions_for_ajax
    def patch(self, request, project_id, observation_id):
        """
        Updates and returns a single contribution

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the request
        project_id : int
            identifies the project in the data base
        observation_id : int
            identifies the observation in the data base

        Returns
        -------
        rest_framework.response.Respone
            Contains the updated serialised observation
        """
        contribution = self.get_contribution(
            request.user,
            project_id,
            observation_id
        )

        return self.update_and_respond(request, contribution)

    @handle_exceptions_for_ajax
    def delete(self, request, project_id, observation_id):
        """
        Deletes a single contribution

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the request
        project_id : int
            identifies the project in the data base
        observation_id : int
            identifies the observation in the data base

        Returns
        -------
        rest_framework.response.Respone
            Empty response indicating successful delete
        """
        contribution = self.get_contribution(
            request.user,
            project_id,
            observation_id
        )

        return self.delete_and_respond(request, contribution)



"""Views for users."""

from json import loads as json_loads
from django.views.generic import TemplateView, CreateView
from django.shortcuts import redirect
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.utils.html import strip_tags
from django.utils.safestring import mark_safe

from braces.views import LoginRequiredMixin

from allauth.account.models import EmailAddress
from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.models import SocialAccount

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status

from geokey.applications.models import Application

from geokey.projects.views import ProjectContext
from geokey.core.decorators import (
    handle_exceptions_for_ajax, handle_exceptions_for_admin
)
from geokey.projects.models import Project
from geokey.projects.base import STATUS

from .models import User
from .serializers import (UserSerializer, UserGroupSerializer)
from .forms import (
    UsergroupCreateForm,
    CustomPasswordChangeForm,
    UserRegistrationForm,
    UserForm
)


# ############################################################################
#
# ADMIN VIEWS
#
# ############################################################################


class Index(TemplateView):
    """
    Displays the splash page. Redirects to dashboard if a user is looged in.
    """
    template_name = 'index.html'

    def get(self, request, *args, **kwargs):
        """
        Renders the page or redirects to dashboard if user is authenticated.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request.

        Returns
        -------
        django.http.HttpResponse
            Rendered template

        django.http.HttpResponseRedirect
            to dashboard, if the user is signed in
        """
        if not request.user.is_anonymous():
            return redirect('admin:dashboard')

        return super(Index, self).get(request, *args, **kwargs)


class Dashboard(LoginRequiredMixin, TemplateView):
    """Dashboard page."""

    template_name = 'dashboard.html'

    def get_context_data(self):
        """
        Return the context to render the view.

        Overwrite the method to add projects, project status types and
        extensions to the context.

        Returns
        -------
        dict
        """
        from geokey.extensions.base import extensions

        ext = []
        for ext_id in sorted(extensions):
            extension = extensions.get(ext_id)
            if extension.get('display_admin') and (
                    not extension.get('superuser') or
                    self.request.user.is_superuser):
                ext.append(extension)

        return {
            'projects': Project.objects.get_list(
                self.request.user).filter(admins=self.request.user),
            'status_types': STATUS,
            'extensions': ext
        }


class UserGroupList(LoginRequiredMixin, ProjectContext, TemplateView):

    """
    Displays the list of user groups in the project.
    """
    template_name = 'users/usergroup_list.html'


class UserGroupCreate(LoginRequiredMixin, ProjectContext, CreateView):

    """
    Provides the form to create a new user group.
    """
    template_name = 'users/usergroup_create.html'
    form_class = UsergroupCreateForm

    @handle_exceptions_for_admin
    def get_context_data(self, **kwargs):
        """
        Overwrites the method to add project to the context.

        Returns
        -------
        dict
            Context
        """

        project_id = self.kwargs['project_id']
        return super(UserGroupCreate, self).get_context_data(project_id)

    def get_success_url(self):
        """
        Returns the redirect URL that is called after the user group has been
        created.

        Returns
        -------
        str
            URL that is called after the Group has been created
        """

        project_id = self.kwargs['project_id']
        return reverse(
            'admin:usergroup_overview',
            kwargs={'project_id': project_id, 'usergroup_id': self.object.id}
        )

    def form_valid(self, form):
        """
        Creates the user group.

        Parameter
        ---------
        form : geokey.users.forms.UsergroupCreateForm
            Represents the user input
        """

        project_id = self.kwargs['project_id']
        project = Project.objects.as_admin(self.request.user, project_id)

        form.instance.project = project

        add_another_url = reverse(
            'admin:usergroup_create',
            kwargs={
                'project_id': project_id
            }
        )

        messages.success(
            self.request,
            mark_safe('The user group has been created. <a href="%s">Add '
                      'another user group.</a>' % add_another_url)
        )

        return super(UserGroupCreate, self).form_valid(form)

    def post(self, request, project_id, *args, **kwargs):
        """
        Creates the user group.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request.
        project_id : int
            identifies the project in the data base

        Returns
        -------
        django.http.HttpResponse
            Rendered template
        """

        self.object = None
        context = self.get_context_data(*args, **kwargs)
        project = context.get('project')

        if project is not None:
            cannot_create = 'New user groups cannot be created.'

            if project.islocked:
                messages.error(
                    self.request,
                    'The project is locked. %s' % cannot_create
                )
            else:
                return super(UserGroupCreate, self).post(
                    request, project_id, *args, **kwargs
                )

        return self.render_to_response(context)


class AdministratorsOverview(LoginRequiredMixin, ProjectContext, TemplateView):

    """
    Displays the project admins overview page.
    """
    template_name = 'users/usergroup_admins.html'


class UserGroupContext(object):

    """
    Mixin that provides the context to render templates. The context contains
    a user group instance based on project_id and subset_id.
    """

    @handle_exceptions_for_admin
    def get_context_data(self, project_id, usergroup_id, *args, **kwargs):
        """
        Returns the context containing the project and user group instances,
        also status types.

        Parameters
        ----------
        project_id : int
            Identifies the project in the database
        usergroup_id : int
            Identifies the user group in the database

        Returns
        -------
        dict
            Context
        """

        project = Project.objects.as_admin(self.request.user, project_id)
        usergroup = project.usergroups.get(pk=usergroup_id)

        return super(UserGroupContext, self).get_context_data(
            project=project,
            usergroup=usergroup,
            status_types=STATUS,
            *args,
            **kwargs
        )


class UserGroupOverview(LoginRequiredMixin, UserGroupContext, TemplateView):

    """
    Displays the user group overview (members) page.
    """
    template_name = 'users/usergroup_overview.html'


class UserGroupSettings(LoginRequiredMixin, UserGroupContext, TemplateView):

    """
    Provides the form to update the user group settings.
    """
    template_name = 'users/usergroup_settings.html'

    def post(self, request, project_id, usergroup_id):
        """
        Updates the user group based on the data entered by the user.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database
        usergroup_id : int
            Identifies the user group in the database

        Returns
        -------
        django.http.HttpResponse
            Rendered template when user group updated
        django.http.HttpResponse
            Rendered template, if project or user group does not exist
        """

        data = request.POST
        context = self.get_context_data(project_id, usergroup_id)
        usergroup = context.get('usergroup')

        if usergroup:
            usergroup.name = strip_tags(data.get('name'))
            usergroup.description = strip_tags(data.get('description'))
            usergroup.save()

            messages.success(self.request, 'The user group has been updated.')

        return self.render_to_response(context)


class UserGroupData(LoginRequiredMixin, UserGroupContext, TemplateView):

    """
    Provides the form to change the filter settings for the user group.
    """
    template_name = 'users/usergroup_data.html'

    def post(self, request, project_id, usergroup_id):
        """
        Updates the user group filter based on the data entered by the user.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database
        usergroup_id : int
            Identifies the user group in the database

        Returns
        -------
        django.http.HttpResponse
            Rendered template when user group updated
        django.http.HttpResponse
            Rendered template, if project or user group does not exist
        """

        data = request.POST
        context = self.get_context_data(project_id, usergroup_id)
        usergroup = context.get('usergroup')

        if usergroup:
            cannot_modify = 'User group data cannot be modified.'

            if usergroup.project.islocked:
                messages.error(
                    self.request,
                    'The project is locked. %s' % cannot_modify
                )
            elif usergroup.project.categories.count() == 0:
                messages.error(
                    self.request,
                    'The project has no categories. %s' % cannot_modify
                )
            else:
                if data['filters'] != '-1':
                    if data['permission'] == 'all':
                        usergroup.filters = None
                    else:
                        usergroup.filters = json_loads(data['filters'])

                    usergroup.save()

                    messages.success(
                        self.request,
                        'The user group has been updated.'
                    )

        return self.render_to_response(context)


class UserGroupPermissions(LoginRequiredMixin, UserGroupContext, TemplateView):

    """
    Provides the form to update the user group permissions.
    """
    template_name = 'users/usergroup_permissions.html'

    def post(self, request, project_id, usergroup_id):
        """
        Updates the user group permissions based on the data entered by the
        user.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database
        usergroup_id : int
            Identifies the user group in the database

        Returns
        -------
        django.http.HttpResponse
            Rendered template when user group updated
        django.http.HttpResponse
            Rendered template, if project or user group does not exist
        """

        data = request.POST
        context = self.get_context_data(project_id, usergroup_id)
        usergroup = context.get('usergroup')

        if usergroup:
            cannot_modify = 'User group permissions cannot be modified.'

            if usergroup.project.islocked:
                messages.error(
                    self.request,
                    'The project is locked. %s' % cannot_modify
                )
            else:
                usergroup.can_moderate = False
                usergroup.can_contribute = False

                if data['permission'] == 'can_moderate':
                    usergroup.can_moderate = True
                elif data['permission'] == 'can_contribute':
                    usergroup.can_contribute = True

                usergroup.save()

                messages.success(
                    self.request,
                    'The user group has been updated.'
                )

        return self.render_to_response(context)


class UserGroupDelete(LoginRequiredMixin, UserGroupContext, TemplateView):

    """
    Deletes the user group.
    """
    template_name = 'base.html'

    def get(self, request, project_id, usergroup_id):
        """
        Deletes the user group.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database
        usergroup_id : int
            Identifies the user group in the database

        Returns
        -------
        django.http.HttpResponseRedirect
            Redirects to user group list if user group is deleted, user group
            settings if project is locked
        django.http.HttpResponse
            Rendered template, if project or user group does not exist
        """

        context = self.get_context_data(project_id, usergroup_id)
        usergroup = context.get('usergroup')

        if usergroup:
            if usergroup.project.islocked:
                messages.error(
                    self.request,
                    'The project is locked. User group cannot be deleted.'
                )
                return redirect(
                    'admin:usergroup_settings',
                    project_id=project_id,
                    usergroup_id=usergroup_id
                )
            else:
                usergroup.delete()

                messages.success(
                    self.request,
                    'The user group has been deleted.'
                )

                return redirect('admin:usergroup_list', project_id=project_id)

        return self.render_to_response(context)


class UserProfile(LoginRequiredMixin, TemplateView):
    """User profile page."""

    template_name = 'users/profile.html'

    def get_context_data(self, *args, **kwargs):
        """
        Return the context to render the view.

        Add social accounts for user to the context.

        Returns
        -------
        dict
        """
        return super(UserProfile, self).get_context_data(
            accounts=SocialAccount.objects.filter(user=self.request.user),
            *args,
            **kwargs
        )

    def post(self, request):
        """
        Update user profile.

        Parameters
        ----------
        request : django.http.HttpRequest
            Object representing the request.

        Returns
        -------
        django.http.HttpResponse
            Rendered template.
        """
        user = User.objects.get(pk=request.user.pk)
        form = UserForm(request.POST, instance=user)

        if form.is_valid():
            if form.has_changed():
                user.display_name = form.cleaned_data['display_name']
                user.email = form.cleaned_data['email']
                user.save()

                if user.email != request.user.email:
                    try:
                        EmailAddress.objects.get(user=user).change(
                            request,
                            user.email,
                            confirm=True
                        )
                    except EmailAddress.DoesNotExist:
                        EmailAddress.objects.create(
                            user=user,
                            email=user.email
                        ).send_confirmation(request)

                messages.success(request, 'Your profile has been updated.')
                self.request.user = user
            else:
                messages.info(request, 'Your profile has not been edited.')

        return self.render_to_response(self.get_context_data(form=form))


class AccountDisconnect(LoginRequiredMixin, TemplateView):
    """Disconnect an account."""

    template_name = 'base.html'

    def get(self, request, account_id):
        """
        Handle the GET request of the view.

        Disconnect the account.

        Parameters
        ----------
        request : django.http.HttpRequest
            Object representing the request.
        account_id : int
            ID identifying the account in the database.

        Returns
        -------
        django.http.HttpResponseRedirect
            Redirecting to user profile page.
        """
        account = None

        try:
            account = SocialAccount.objects.get(
                pk=account_id,
                user=request.user)
        except:
            messages.error(request, 'The account could not be found.')

        if account:
            try:
                get_adapter(request).validate_disconnect(
                    account,
                    SocialAccount.objects.filter(user=self.request.user))
                account.delete()
                messages.success(request, 'The account has been disconnected.')
            except ValidationError, e:
                messages.error(request, e)

        return HttpResponseRedirect(reverse('admin:userprofile'))


# ############################################################################
#
# AJAX VIEWS
#
# ############################################################################

class QueryUsers(APIView):

    """
    AJAX endpoint for querying a list of users in the AJAX API.
    """

    def get(self, request):
        """
        Returns a list of users where the `display_name` matches the query.

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the HTTP request

        Returns
        -------
        rest_framework.response.Response
            Contains the list of users
        """

        q = request.GET.get('query')

        if q is None:
            return Response([])
        else:
            users = User.objects.filter(
                display_name__icontains=q.lower()
            ).exclude(pk=1)[:10]

            serializer = UserSerializer(
                users, many=True, fields=('id', 'display_name')
            )
            return Response(serializer.data)


class UserGroup(APIView):

    """
    API endpoints for a user group of a project in the AJAX API.
    """

    @handle_exceptions_for_ajax
    def put(self, request, project_id, usergroup_id):
        """
        Updates user group information.

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the HTTP request
        project_id : int
            Identifies the project in the database
        usergroup_id : int
            Identifies the user group in the database

        Returns
        -------
        rest_framework.response.Response
            Contains the serialised user group or an error message
        """

        project = Project.objects.as_admin(request.user, project_id)

        if project.islocked:
            return Response(
                'The project is locked. User group info cannot be modified.',
                status=status.HTTP_400_BAD_REQUEST
            )
        else:
            group = project.usergroups.get(pk=usergroup_id)
            serializer = UserGroupSerializer(
                group, data=request.data, partial=True)

            if serializer.is_valid():
                serializer.save()
                return Response(serializer.data)

            return Response(
                serializer.errors,
                status=status.HTTP_400_BAD_REQUEST
            )


class UserGroupUsers(APIView):

    """
    API endpoints for all users in a user group of a project.
    """

    @handle_exceptions_for_ajax
    def post(self, request, project_id, usergroup_id):
        """
        Adds a user to the user group.

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the HTTP request
        project_id : int
            Identifies the project in the database
        usergroup_id : int
            Identifies the user group in the database

        Returns
        -------
        rest_framework.response.Response
            Contains the serialised user group or an error message
        """

        project = Project.objects.as_admin(request.user, project_id)

        if project.islocked:
            return Response(
                'The project is locked. New users cannot be added.',
                status=status.HTTP_400_BAD_REQUEST
            )
        else:
            user_id = request.data.get('user_id')

            try:
                user = User.objects.get(pk=user_id)
            except User.DoesNotExist:
                return Response(
                    'The user you are trying to add to the user group does ' +
                    'not exist.',
                    status=status.HTTP_400_BAD_REQUEST
                )

            group = project.usergroups.get(pk=usergroup_id)
            group.users.add(user)

            serializer = UserGroupSerializer(group)
            return Response(serializer.data, status=status.HTTP_201_CREATED)


class UserGroupSingleUser(APIView):

    """
    API endpoints for a single user in a user group of a project.
    """

    @handle_exceptions_for_ajax
    def delete(self, request, project_id, usergroup_id, user_id):
        """
        Removes a user from the user group.

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the HTTP request
        project_id : int
            Identifies the project in the database
        usergroup_id : int
            Identifies the user group in the database
        user_id : int
            Identifies the user in the database

        Returns
        -------
        rest_framework.response.Response
            Empty response if successful or an error message
        """

        project = Project.objects.as_admin(request.user, project_id)

        if project.islocked:
            return Response(
                'The project is locked. Users cannot be removed.',
                status=status.HTTP_400_BAD_REQUEST
            )
        else:
            usergroup = project.usergroups.get(pk=usergroup_id)
            user = usergroup.users.get(pk=user_id)
            usergroup.users.remove(user)
            return Response(status=status.HTTP_204_NO_CONTENT)


# ############################################################################
#
# PUBLIC API VIEWS
#
# ############################################################################


class CreateUserMixin(object):

    """
    Mixin to create a user.
    """

    def create_user(self, data):
        """
        Creates a user.

        Parameters
        ----------
        data : dict
            User data

        Returns
        -------
        geokey.users.models.User
            The newly created user
        """

        user = User.objects.create_user(
            data.get('email'),
            data.get('display_name'),
            password=data.get('password1')
        )
        user.save()

        return user


class UserAPIView(CreateUserMixin, APIView):

    """
    API endpoint to get, update and create a user.
    """

    def get(self, request):
        """
        Returns the user information.

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the HTTP request

        Response
        --------
        rest_framework.response.Response
            Containing the user info or an error message if no user is signed
            in
        """

        user = request.user

        if not user.is_anonymous():
            serializer = UserSerializer(user)

            return Response(
                serializer.data,
                status=status.HTTP_200_OK
            )
        else:
            return Response(
                {'error': 'You have to be signed in to get user information'},
                status=status.HTTP_401_UNAUTHORIZED
            )

    def patch(self, request):
        """
        Updates the user information.

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the HTTP request

        Response
        --------
        rest_framework.response.Response
            Containing the user info or an error message
        """

        user = request.user

        if user.is_anonymous():
            return Response(
                {'error': 'You have to be signed in to get user information'},
                status=status.HTTP_401_UNAUTHORIZED
            )

        data = request.data

        serializer = UserSerializer(user, data=data, partial=True)

        if serializer.is_valid():
            new_mail = data.get('email')

            if new_mail is not None and user.email != new_mail:
                try:
                    email = EmailAddress.objects.get(
                        user=user, email=user.email
                    )
                    email.change(request, new_mail, confirm=True)
                except EmailAddress.DoesNotExist:
                    EmailAddress.objects.create(user=user, email=new_mail)

            serializer.save()

            return Response(
                serializer.data,
                status=status.HTTP_200_OK
            )
        else:
            return Response(
                {'error': serializer.errors},
                status=status.HTTP_400_BAD_REQUEST
            )

    def post(self, request):
        """
        Create a new user.

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the HTTP request

        Response
        --------
        rest_framework.response.Response
            Containing the user info or an error message
        """
        data = request.data
        form = UserRegistrationForm(data)

        try:
            Application.objects.get(client_id=data.get('client_id'))

            if form.is_valid():
                user = self.create_user(data)

                EmailAddress.objects.add_email(
                    request._request,
                    user,
                    user.email,
                    signup=True,
                    confirm=True
                )

                serializer = UserSerializer(user)

                return Response(
                    serializer.data,
                    status=status.HTTP_201_CREATED
                )
            else:
                return Response(
                    {'errors': form.errors},
                    status=status.HTTP_400_BAD_REQUEST
                )
        except Application.DoesNotExist:
            return Response(
                {'errors': {'client': 'Client ID not provided or incorrect.'}},
                status=status.HTTP_400_BAD_REQUEST
            )


class ChangePasswordView(APIView):

    """
    API endpoint to change the password via the client.
    """

    def post(self, request):
        """
        Updates the password.

        Parameters
        ----------
        request : rest_framework.request.Request
            Represents the HTTP request

        Response
        --------
        rest_framework.response.Response
            Empty response indicating success or error message
        """

        user = request.user
        data = request.data

        if not user.is_anonymous():
            form = CustomPasswordChangeForm(user, data)

            if form.is_valid():
                form.save()
                return Response(status=status.HTTP_204_NO_CONTENT)
            else:
                return Response(
                    {'errors': form.errors},
                    status=status.HTTP_400_BAD_REQUEST
                )
        else:
            return Response(
                {'error': 'You have to be signed in to change the password.'},
                status=status.HTTP_401_UNAUTHORIZED
            )



"""Tests for serializers of users."""

from django.test import TestCase

from rest_framework.serializers import ValidationError
from allauth.socialaccount.models import SocialAccount

from .model_factories import UserFactory
from ..serializers import UserSerializer


class UserSerializerTest(TestCase):

    def test_get_social_accounts_when_empty(self):
        user = UserFactory.create()
        serializer = UserSerializer(user)
        self.assertEquals(serializer.get_social_accounts(user), [])

    def test_get_social_accounts_when_connected_to_facebook(self):
        user = UserFactory.create()
        social_account = SocialAccount.objects.create(
            user=user, provider='facebook', uid='123')
        serializer = UserSerializer(user)
        self.assertEquals(
            serializer.get_social_accounts(user),
            [{
                'id': social_account.id,
                'provider': 'facebook'
            }]
        )

    def test_validate_display_name(self):
        UserFactory.create(**{'display_name': 'name'})
        user = UserFactory.create()
        serializer = UserSerializer(user)

        try:
            serializer.validate_display_name('name')
        except ValidationError:
            pass
        else:
            self.fail('validate_display_name did not raise ValidationError')

    def test_validate_email(self):
        UserFactory.create(**{'email': 'name@example.com'})
        user = UserFactory.create()
        serializer = UserSerializer(user)

        try:
            serializer.validate_email('name@example.com')
        except ValidationError:
            pass
        else:
            self.fail('validate_email did not raise ValidationError')



"""Tests for base of extensions."""

from django.test import TestCase

from geokey.extensions.base import extensions, register, deregister
from geokey.extensions.exceptions import ExtensionExists


class RegisterTest(TestCase):
    """Test register."""

    def tearDown(self):
        """Tear down test."""
        deregister(self.ext_id)

    def test_register(self):
        """Test registering new extension."""
        self.ext_id = 'test_ext'
        register(self.ext_id, 'Test', True, True, '1.0.0')

        extension = extensions.get(self.ext_id)
        self.assertEqual(extension.get('ext_id'), self.ext_id)
        self.assertEqual(extension.get('name'), 'Test')
        self.assertEqual(extension.get('version'), '1.0.0')
        self.assertTrue(extension.get('display_admin'))
        self.assertTrue(extension.get('superuser'))
        self.assertEqual(extension.get('index_url'), self.ext_id + ':index')

    def test_register_when_already_exists(self):
        """Test registering existing extension."""
        self.ext_id = 'test_ext'
        extensions[self.ext_id] = {
            'ext_id': self.ext_id,
            'name': 'Test',
            'version': '1.0.0',
            'display_admin': True,
            'superuser': True,
            'index_url': self.ext_id + ':index'
        }

        with self.assertRaises(ExtensionExists):
            register(self.ext_id, 'Test B', False, False, '1.0.0')


class DeregisterTest(TestCase):
    """Test deregister."""

    def test_deregister(self):
        """Test deregistering existing extension."""
        ext_id = 'test_ext'
        extensions[ext_id] = {
            'ext_id': ext_id,
            'name': 'Test',
            'version': '1.0.0',
            'display_admin': True,
            'superuser': True,
            'index_url': ext_id + ':index'
        }
        deregister(ext_id)

        self.assertNotIn(ext_id, extensions)



"""Models for contributions."""

import re

from pytz import utc
from datetime import datetime
from iso8601 import parse_date
from iso8601.iso8601 import ParseError

from django.db import models
from django.conf import settings
from django.core.exceptions import ValidationError
from django.dispatch import receiver
from django.db.models.signals import pre_save, post_save
from django.contrib.gis.db import models as gis

from django_pgjson.fields import JsonBField
from simple_history.models import HistoricalRecords

from geokey.core.exceptions import InputError

from .base import (
    OBSERVATION_STATUS,
    COMMENT_STATUS,
    COMMENT_REVIEW,
    LOCATION_STATUS,
    MEDIA_STATUS
)
from .managers import (
    ObservationManager,
    LocationManager,
    CommentManager,
    MediaFileManager
)


class Location(models.Model):
    """
    Represents a location to which an arbitrary number of observations can be
    attached.
    """
    name = models.CharField(max_length=100, blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    geometry = gis.GeometryField(geography=True)
    created_at = models.DateTimeField(auto_now_add=True)
    creator = models.ForeignKey(settings.AUTH_USER_MODEL)
    version = models.IntegerField(default=1)
    private = models.BooleanField(default=False)
    private_for_project = models.ForeignKey('projects.Project', null=True)
    status = models.CharField(
        choices=LOCATION_STATUS,
        default=LOCATION_STATUS.active,
        max_length=20
    )

    objects = LocationManager()


class Observation(models.Model):
    """
    Stores a single observation.
    """
    location = models.ForeignKey(
        Location, related_name='locations'
    )
    project = models.ForeignKey(
        'projects.Project', related_name='observations'
    )
    category = models.ForeignKey('categories.Category')
    status = models.CharField(
        choices=OBSERVATION_STATUS,
        default=OBSERVATION_STATUS.active,
        max_length=20
    )
    properties = JsonBField(default={})
    created_at = models.DateTimeField(auto_now_add=True)
    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='creator'
    )
    updated_at = models.DateTimeField(null=True, auto_now_add=True)
    updator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='updator',
        null=True
    )
    version = models.IntegerField(default=1)
    search_index = models.TextField(null=True, blank=True)
    display_field = models.TextField(null=True, blank=True)
    expiry_field = models.DateTimeField(null=True, blank=True)
    num_media = models.IntegerField(default=0)
    num_comments = models.IntegerField(default=0)

    history = HistoricalRecords()
    objects = ObservationManager()

    class Meta:
        ordering = ['-updated_at', 'id']

    @classmethod
    def validate_partial(cls, category, data):
        """
        Validates the data against the category field definition. This is a
        partial validation, which is used to validate drafts, field values
        that are not provided are not validated.

        Parameter
        ---------
        category : geokey.categories.models.Category
            Category that the data is validated against
        data : dict
            Dictionary of key-value-pairs; incoming data that is validated

        Raises
        ------
        ValidationError:
            when data is invalid
        """
        is_valid = True
        error_messages = []

        for field in category.fields.all().filter(status='active'):
            if field.key in data and data.get(field.key) is not None:
                try:
                    field.validate_input(data.get(field.key))
                except InputError, error:
                    is_valid = False
                    error_messages.append(error)

        if not is_valid:
            raise ValidationError(error_messages)

    @classmethod
    def validate_full(cls, category, data):
        """
        Validates the data against the category field definition. This is a
        full validation.

        Parameter
        ---------
        category : geokey.categories.models.Category
            Category that the data is validated against
        data : dict
            Dictionary of key-value-pairs; incoming data that is validated

        Raises
        ------
        ValidationError:
            when data is invalid
        """
        is_valid = True
        error_messages = []

        for field in category.fields.all().filter(status='active'):
            try:
                field.validate_input(data.get(field.key))
            except InputError, error:
                is_valid = False
                error_messages.append(error)

        if not is_valid:
            raise ValidationError(error_messages)

    @classmethod
    def create(cls, properties=None, creator=None, location=None,
               category=None, project=None, status=None):
        """
        Creates and returns a new observation. Validates all fields first and
        raises a ValidationError if at least one field did not validate.
        Creates the object if all fields are valid.

        Parameter
        ---------
        properties : dict
            Attributes of the observation
        creator : geokey.users.models.User
            User who creates the observation
        location : geokey.contributions.models.Location
            Location of the contribution
        category : geokey.categories.models.Category
            Category of the contribution
        project : geokey.projects.models.Project
            Project the contribution is assigned to
        status : str
            Status of the contribution; one of active, review, pending or draft

        Return
        ------
        geokey.contributions.models.Observation
            The observation created
        """
        if not properties:
            properties = {}

        location.save()
        observation = cls.objects.create(
            location=location,
            category=category,
            project=project,
            properties=properties,
            creator=creator,
            status=status
        )
        return observation

    def update(self, properties, updator, status=None):
        """
        Updates data of the observation

        Parameter
        ---------
        properties : dict
            Attributes of the observation
        updator : geokey.users.models.User
            User who creates the observation
        status : str
            Status of the contribution; one of active, review, pending or draft

        Return
        ------
        geokey.contributions.models.Observation
            The updated observation
        """
        if status != 'draft':
            self.version = self.version + 1

        if properties:
            self.properties = properties
        self.updator = updator
        self.status = status or self.status
        self.updated_at = datetime.utcnow().replace(tzinfo=utc)

        self.save()
        return self

    def update_display_field(self):
        """
        Updates the display_field attribute. It uses the display field of the
        contributions category and adds a string line 'key:value' to the
        display field property
        """
        display_field = self.category.display_field
        value = None

        if display_field:
            if self.properties:
                value = self.properties.get(display_field.key)

            self.display_field = '%s:%s' % (display_field.key, value)

    def update_expiry_field(self):
        """
        Updates the expiry_field attribute. It uses the expiry field of the
        contributions category and sets the date according to the value set
        for the current contribution.
        """
        expiry_field = self.category.expiry_field
        value = None

        if expiry_field and self.properties:
            try:
                value = parse_date(self.properties.get(expiry_field.key))
            except ParseError:
                pass

        self.expiry_field = value

    def update_count(self):
        """
        Updates the count of media files attached and comments. Should be
        called each time a file or comment is added/deleted.
        """
        self.num_media = self.files_attached.count()
        self.num_comments = self.comments.count()
        self.save()

    def create_search_index(self):
        search_index = []

        for field in self.category.fields.all():
            value = None
            if self.properties and field.key in self.properties.keys():
                if field.fieldtype == 'TextField':
                    value = self.properties.get(field.key)

                if field.fieldtype == 'NumericField':
                    value = str(self.properties.get(field.key))

                if field.fieldtype == 'LookupField':
                    lookup_id = self.properties.get(field.key)
                    if lookup_id:
                        value = field.lookupvalues.get(pk=lookup_id).name

                if field.fieldtype == 'MultipleLookupField':
                    lookup_id = self.properties.get(field.key)
                    if lookup_id:
                        lookupvalues = field.lookupvalues.filter(
                            pk__in=lookup_id
                        )

                        value = ' '.join([val.name for val in lookupvalues])

            if value:
                cleaned = re.sub(r'[\W_]+', ' ', value)
                terms = cleaned.lower().split()

                search_index = search_index + list(
                    set(terms) - set(search_index)
                )

        self.search_index = ','.join(search_index)

    def delete(self):
        """
        Deletes the observation by setting it's status to DELETED
        """
        self.status = OBSERVATION_STATUS.deleted
        self.save()


@receiver(pre_save, sender=Observation)
def pre_save_observation_update(sender, **kwargs):
    """
    Receiver that is called before an observation is saved. Updates
    `search_index`, `display_field`, `expiry_field` properties.
    """
    observation = kwargs.get('instance')
    observation.update_display_field()
    observation.update_expiry_field()
    observation.create_search_index()


class Comment(models.Model):
    """
    A comment that is added to a contribution.
    """
    text = models.TextField()
    created_at = models.DateTimeField(auto_now_add=True)
    creator = models.ForeignKey(settings.AUTH_USER_MODEL)
    commentto = models.ForeignKey('Observation', related_name='comments')
    respondsto = models.ForeignKey(
        'Comment',
        null=True,
        blank=True,
        related_name='responses'
    )
    status = models.CharField(
        choices=COMMENT_STATUS,
        default=COMMENT_STATUS.active,
        max_length=20
    )
    review_status = models.CharField(
        choices=COMMENT_REVIEW,
        null=True,
        blank=True,
        max_length=10
    )

    objects = CommentManager()
    history = HistoricalRecords()

    class Meta:
        ordering = ['id']

    def delete(self):
        """
        Deletes the comment by setting it's status to DELETED
        """
        self.responses.all().delete()
        self.status = COMMENT_STATUS.deleted
        self.save()


@receiver(post_save, sender=Comment)
def post_save_comment_count_update(sender, **kwargs):
    """
    Receiver that is called after a comment is saved. Updates num_media and
    num_comments properties.
    """
    comment = kwargs.get('instance')
    comment.commentto.update_count()


class MediaFile(models.Model):
    """
    Base class for all media files. Not to be instaciate; instaciate one of
    the child classes instead.
    """
    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    contribution = models.ForeignKey(
        'contributions.Observation', related_name='files_attached'
    )
    creator = models.ForeignKey(settings.AUTH_USER_MODEL)
    created_at = models.DateTimeField(auto_now_add=True)
    status = models.CharField(
        choices=MEDIA_STATUS,
        default=MEDIA_STATUS.active,
        max_length=20
    )

    objects = MediaFileManager()

    class Meta:
        ordering = ['id']

    @property
    def type_name(self):
        """
        Returns the type of media file. To be implemented by child classes.

        Raises
        ------
        NotImplementedError
            if called on MediaFile base class
        """
        raise NotImplementedError(
            'The property `type_name` has not been implemented for this '
            'subclass of `MediaFile`.'
        )

    def delete(self):
        """
        Deletes a file by setting its status to deleted
        """
        self.status = MEDIA_STATUS.deleted
        self.save()


class AudioFile(MediaFile):
    """
    Stores audio files uploaded by users.
    """
    audio = models.FileField(upload_to='user-uploads/audio')

    class Meta:
        ordering = ['id']
        app_label = 'contributions'

    @property
    def type_name(self):
        """
        Returns file type name

        Returns
        -------
        str
            'AudioFile'
        """
        return 'AudioFile'


class ImageFile(MediaFile):
    """
    Stores images uploaded by users.
    """
    image = models.ImageField(upload_to='user-uploads/images')

    class Meta:
        ordering = ['id']
        app_label = 'contributions'

    @property
    def type_name(self):
        """
        Returns file type name

        Returns
        -------
        str
            'ImageFile'
        """
        return 'ImageFile'


class VideoFile(MediaFile):
    """
    Stores images uploaded by users.
    """
    video = models.ImageField(upload_to='user-uploads/videos')
    youtube_id = models.CharField(max_length=100)
    thumbnail = models.ImageField(upload_to='user-uploads/videos', null=True)
    youtube_link = models.URLField(max_length=255, null=True, blank=True)
    swf_link = models.URLField(max_length=255, null=True, blank=True)

    class Meta:
        ordering = ['id']

    @property
    def type_name(self):
        """
        Returns file type name

        Returns
        -------
        str
            'VideoFile'
        """
        return 'VideoFile'


@receiver(post_save)
def post_save_media_file_count_update(sender, **kwargs):
    """
    Receiver that is called after a media file is saved. Updates num_media and
    num_comments properties.
    """
    if sender.__name__ in ['ImageFile', 'VideoFile', 'AudioFile']:
        media_file = kwargs.get('instance')
        media_file.contribution.update_count()



"""Models for subsets."""

from django.conf import settings
from django.db import models

from django_pgjson.fields import JsonBField
from simple_history.models import HistoricalRecords


from geokey.core.mixins import FilterMixin


class Subset(FilterMixin, models.Model):
    """Stores a single subset."""

    name = models.CharField(max_length=100)
    description = models.TextField(blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    creator = models.ForeignKey(settings.AUTH_USER_MODEL)
    project = models.ForeignKey('projects.Project', related_name='subsets')
    filters = JsonBField(blank=True, null=True)
    where_clause = models.TextField(blank=True, null=True)
    history = HistoricalRecords()



"""Tests for template tags of contributions (observations)."""

from django.test import TestCase

from geokey.contributions.templatetags import kml_tags

from geokey.categories.tests.model_factories import CategoryFactory


class TemplateTagsTest(TestCase):
    def test_kml_name(self):
        name = kml_tags.kml_name({'display_field': {'value': 'Awesome pub'}})
        self.assertEqual(name, 'Awesome pub')

    def test_kml_name_when_no_display_field(self):
        name = kml_tags.kml_name(None)
        self.assertEqual(name, '')

    def test_kml_name_when_display_field_is_empty(self):
        name = kml_tags.kml_name({'display_field': {}})
        self.assertEqual(name, '')

    def test_geom(self):
        data = {
            "location": {
                "geometry": {"type": "Point", "coordinates": [-2.812, 51.179]}
            }
        }
        kml_geom = kml_tags.kml_geom(data)
        self.assertEqual(
            kml_geom,
            '<Point><coordinates>-2.812,51.179</coordinates></Point>'
        )

    def test_description(self):
        CategoryFactory.create()
        data = {
            'properties': {
                'key_1': 1,
                'key_2': 'value'
            },
            'meta': {
                'category': {
                    'id': 1,
                }
            },
            'media': [
                {
                    'file_type': 'ImageFile',
                    'name': 'Image',
                    'description': None,
                    'url': '/media/image.png',
                    'thumbnail_url': '/media/image-thumbnail.png'
                },
                {
                    'file_type': 'VideoFile',
                    'name': 'Video',
                    'description': None,
                    'url': 'https://www.youtube.com/watch?v=LueGR1IpfS0',
                    'thumbnail_url': '/static/img/play.png'
                },
                {
                    'file_type': 'AudioFile',
                    'name': 'Audio',
                    'description': 'New feature - audio support.',
                    'url': '/media/audio.mp4',
                    'thumbnail_url': None
                }
            ],
            'comments': [
                {
                    'text': 'Is it ready yet?',
                    'creator': {
                        'display_name': 'Adam'
                    },
                    'responses': [
                        {
                            'text': 'Yes!',
                            'creator': {
                                'display_name': 'Sarah'
                            },
                            'responses': []
                        }
                    ]
                }
            ]
        }
        description = kml_tags.kml_desc(data)
        self.assertEqual(
            description,
            '<![CDATA[<table><tr><td>key_1</td><td>1</td></tr>'
            '<tr><td>key_2</td><td>value</td></tr></table>'
            '<table><tr><td><strong>Image</strong><br /><a href="/media/image.png"><img src="/media/image-thumbnail.png" /></a></td></tr>'
            '<tr><td><strong>Video</strong><br /><a href="https://www.youtube.com/watch?v=LueGR1IpfS0"><img src="/static/img/play.png" /></a></td></tr>'
            '<tr><td><a href="/media/audio.mp4"><strong>Audio</strong></a><br />New feature - audio support.</td></tr></table>'
            '<table><tr><td><strong>Adam</strong><br />Is it ready yet?'
            '<table><tr><td><strong>Sarah</strong><br />Yes!</td></tr></table></td></tr></table>]]>'
        )

    def test_style(self):
        data = {
            'meta': {
                'category': {
                    'colour': '#ff0000'
                }
            }
        }
        colour = kml_tags.kml_style(data)
        self.assertEqual(colour, 'ff0000')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('users', '0002_auto_20150904_1113'),
        ('users', '0005_auto_20150825_0933'),
    ]

    operations = [
    ]



"""Template tags for filtering fields."""

from django import template


register = template.Library()


@register.filter
def only_fields(fields, type_names):
    type_names = [type_name.strip() for type_name in type_names.split(',')]
    return [field for field in fields if field.type_name in type_names]


@register.filter
def except_fields(fields, type_names):
    type_names = [type_name.strip() for type_name in type_names.split(',')]
    return [field for field in fields if field.type_name not in type_names]



"""Base for extensions."""

from geokey.extensions.exceptions import ExtensionExists


extensions = {}


def register(ext_id, name, display_admin=False, superuser=False, version=None):
    """
    Register a new extension on the system.

    Parameters
    ----------
    ext_id : str
        Unique identifier for the extension.
    name : str
        Human readable name of the extension.
    display_admin : bool
        Indicates if the extension provides pages for the admin interface.
    superuser : bool
        Indicates if the extentsion is available for superusers only.
    version : str
        Version of the extension (optional).

    Raises
    ------
    ExtensionExists
        When another extension with the same `ext_id` has already been
        registered.
    """
    if ext_id in extensions.keys():
        raise ExtensionExists(
            'An extension with ID %s has already been registered.' % ext_id
        )

    extensions[ext_id] = {
        'ext_id': ext_id,
        'name': name,
        'version': version,
        'display_admin': display_admin,
        'superuser': superuser,
        'index_url': ext_id + ':index'
    }


def deregister(ext_id):
    """
    Deregister an extension from the system.

    Only to be used for testing.

    Parameters
    ----------
    ext_id : str
        Unique identifier for the extension.
    """
    extensions.pop(ext_id, None)



"""Mixins for categories."""

from geokey.core.decorators import handle_exceptions_for_admin
from geokey.categories.models import Category, Field


class CategoryMixin(object):
    """A mixin for category."""

    @handle_exceptions_for_admin
    def get_context_data(self, project_id, category_id, *args, **kwargs):
        """
        Return the context to render the view.

        Overwrite the method to add the project and the category to the
        context.

        Returns
        -------
        dict
            Context.
        """
        category = Category.objects.as_admin(
            self.request.user,
            project_id,
            category_id
        )
        return super(CategoryMixin, self).get_context_data(
            project=category.project,
            category=category,
            *args,
            **kwargs
        )


class FieldMixin(object):
    """A mixin for field."""

    @handle_exceptions_for_admin
    def get_context_data(self, project_id, category_id, field_id,
                         *args, **kwargs):
        """
        Return the context to render the view.

        Overwrite the method to add the project, the category and the field to
        the context.

        Returns
        -------
        dict
            Context.
        """
        field = Field.objects.as_admin(
            self.request.user,
            project_id,
            category_id,
            field_id
        )
        return super(FieldMixin, self).get_context_data(
            project=field.category.project,
            category=field.category,
            field=field,
            is_display_field=(field == field.category.display_field),
            is_expiry_field=(field == field.category.expiry_field),
            *args,
            **kwargs
        )



"""Production settings."""

from .base import *

# Hosts/domain names that are valid for this site; required if DEBUG is False
# See https://docs.djangoproject.com/en/1.5/ref/settings/#allowed-hosts
ALLOWED_HOSTS = ['*']



"""Core mixins."""


class FilterMixin(object):
    """A mixin for filter."""

    def remove_filter_field(self, field):
        """
        Remove a field from the filter.

        Parameters
        ----------
        field : geokey.categories.models.Field
            Represents the field of a category.
        """
        if self.filters:
            category_filter = self.filters.get(str(field.category.id), None)

            if category_filter:
                field_filter = category_filter.pop(field.key, None)

                if field_filter:
                    self.save()

    def save(self, *args, **kwargs):
        """Overwrite `save` to implement integrity ensurance."""
        self.where_clause = None

        if self.filters is not None:
            queries = []

            for key in self.filters:
                category = self.project.categories.get(pk=key)
                queries.append(category.get_query(self.filters[key]))

            if len(queries) > 0:
                query = ' OR '.join(queries)
                self.where_clause = query
            else:
                self.where_clause = 'FALSE'

        super(FilterMixin, self).save(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import django.contrib.postgres.fields.hstore


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='LoggerHistory',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('user', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('project', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('usergroup', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('category', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('field', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('location', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('observation', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('comment', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('mediafile', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('subset', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('action', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
                ('historical', django.contrib.postgres.fields.hstore.HStoreField(null=True, blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('categories', '0013_auto_20150130_1440'),
    ]

    operations = [
        migrations.AlterField(
            model_name='category',
            name='symbol',
            field=models.ImageField(max_length=500, null=True, upload_to=b'symbols'),
        ),
    ]



"""Tests for template tags of categories."""

from django.test import TestCase

from geokey.categories.templatetags import filter_fields
from geokey.categories.tests.model_factories import (
    CategoryFactory,
    TextFieldFactory,
    NumericFieldFactory,
    DateTimeFieldFactory,
    DateFieldFactory,
    TimeFieldFactory,
    LookupFieldFactory,
    MultipleLookupFieldFactory
)


class OnlyFieldsTest(TestCase):

    def test_only_fields(self):
        category = CategoryFactory()
        TextFieldFactory(category=category)
        NumericFieldFactory(category=category)
        DateTimeFieldFactory(category=category)
        DateFieldFactory(category=category)
        TimeFieldFactory(category=category)
        LookupFieldFactory(category=category)
        MultipleLookupFieldFactory(category=category)

        all_fields = category.fields.all()

        type_names = [
            'Text',
            'Numeric',
            'Date and Time',
            'Date',
            'Time',
            'Select box',
            'Multiple select'
        ]

        for type_name in type_names:
            date_fields = filter_fields.only_fields(all_fields, type_name)
            self.assertEqual(len(date_fields), 1)
            for field in date_fields:
                self.assertEqual(field.type_name, type_name)

        date_fields = filter_fields.only_fields(
            all_fields,
            (', ').join(type_names)
        )
        self.assertEqual(len(date_fields), len(type_names))
        for field in date_fields:
                self.assertTrue(field.type_name in type_names)


class ExceptFieldsTest(TestCase):

    def test_except_fields(self):
        category = CategoryFactory()
        TextFieldFactory(category=category)
        NumericFieldFactory(category=category)
        DateTimeFieldFactory(category=category)
        DateFieldFactory(category=category)
        TimeFieldFactory(category=category)
        LookupFieldFactory(category=category)
        MultipleLookupFieldFactory(category=category)

        all_fields = category.fields.all()

        type_names = [
            'Text',
            'Numeric',
            'Date and Time',
            'Date',
            'Time',
            'Select box',
            'Multiple select'
        ]

        for type_name in type_names:
            date_fields = filter_fields.except_fields(all_fields, type_name)
            self.assertEqual(len(date_fields), len(type_names) - 1)
            for field in date_fields:
                self.assertNotEqual(field.type_name, type_name)

        date_fields = filter_fields.except_fields(
            all_fields,
            (', ').join(type_names)
        )
        self.assertEqual(len(date_fields), 0)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import re


def create_search_index(apps, schema_editor):
    Observation = apps.get_model('contributions', 'Observation')

    for o in Observation.objects.all():
        search_index = []

        fields = o.search_matches.split('#####')
        for field in fields:
            if field:
                value = field.split(':')[1]

                cleaned = re.sub(r'[\W_]+', ' ', value)
                terms = cleaned.lower().split()

                search_index = search_index + list(
                    set(terms) - set(search_index)
                )

        o.search_index = ','.join(search_index)
        o.save()


class Migration(migrations.Migration):

    dependencies = [
        ('contributions', '0013_auto_20150907_1345'),
    ]

    operations = [
        migrations.RunPython(create_search_index)
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('categories', '0016_multiplelookupvalue_symbol'),
    ]

    operations = [
        migrations.AddField(
            model_name='category',
            name='expiry_field',
            field=models.ForeignKey(related_name='expiry_field_of', to='categories.Field', null=True),
        ),
    ]



"""Tests for logger: model Category."""

from django.test import TestCase

from geokey.core.models import LoggerHistory
from geokey.users.tests.model_factories import UserFactory
from geokey.projects.tests.model_factories import ProjectFactory
from geokey.categories.tests.model_factories import CategoryFactory


class LogCategoryTest(TestCase):
    """Test model Category."""

    def setUp(self):
        """Set up test."""
        self.user = UserFactory.create()
        self.project = ProjectFactory.create(**{
            'creator': self.user})
        self.category = CategoryFactory.create(**{
            'creator': self.user,
            'project': self.project})

    def test_log_create(self):
        """Test when category gets created."""
        log_count_init = LoggerHistory.objects.count()
        category = CategoryFactory.create(**{
            'creator': self.user,
            'project': self.project})

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(category.id),
            'name': category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'created',
            'class': 'Category'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_delete(self):
        """Test when category gets deleted."""
        category_id = self.category.id
        category_name = self.category.name
        log_count_init = LoggerHistory.objects.count()
        self.category.delete()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(category_id),
            'name': category_name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'deleted',
            'class': 'Category',
            'field': 'status',
            'value': 'deleted'})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.category.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, category_id)
        self.assertEqual(history.name, category_name)

    def test_log_update_name(self):
        """Test when name changes."""
        log_count_init = LoggerHistory.objects.count()
        original_name = self.category.name
        self.category.name = '%s UPDATED' % self.category.name
        self.category.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Category',
            'field': 'name'})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.category.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.category.id)
        self.assertEqual(history.name, original_name)

    def test_log_update_status(self):
        """Test when status changes."""
        log_count_init = LoggerHistory.objects.count()

        original_status = self.category.status
        self.category.status = 'inactive'
        self.category.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Category',
            'field': 'status',
            'value': self.category.status})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.category.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.category.id)
        self.assertEqual(history.status, original_status)

        original_status = self.category.status
        self.category.status = 'active'
        self.category.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Category',
            'field': 'status',
            'value': self.category.status})
        self.assertEqual(log_count, log_count_init + 2)
        history = self.category.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.category.id)
        self.assertEqual(history.status, original_status)

    def test_log_update_default_status(self):
        """Test when default status changes."""
        log_count_init = LoggerHistory.objects.count()

        original_default_status = self.category.default_status
        self.category.default_status = 'active'
        self.category.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Category',
            'field': 'default_status',
            'value': self.category.default_status})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.category.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.category.id)
        self.assertEqual(history.default_status, original_default_status)

        original_default_status = self.category.default_status
        self.category.default_status = 'pending'
        self.category.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, None)
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Category',
            'field': 'default_status',
            'value': self.category.default_status})
        self.assertEqual(log_count, log_count_init + 2)
        history = self.category.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.category.id)
        self.assertEqual(history.default_status, original_default_status)

    def test_log_update_multiple_fields(self):
        """Test when multiple model fields changes."""
        log_count_init = LoggerHistory.objects.count()
        original_name = self.category.name
        original_status = self.category.status
        self.category.name = '%s UPDATED' % self.category.name
        self.category.status = 'inactive'
        self.category.save()

        log_count = LoggerHistory.objects.count()
        self.assertEqual(log_count, log_count_init + 2)

        logs = LoggerHistory.objects.all().order_by('-pk')[:2]

        # 1st changed field
        self.assertNotEqual(logs[1].user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(logs[1].project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(logs[1].category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(logs[1].field, None)
        self.assertEqual(logs[1].action, {
            'id': 'updated',
            'class': 'Category',
            'field': 'name'})
        history_2 = self.category.history.get(pk=logs[1].historical.get('id'))
        self.assertEqual(history_2.id, self.category.id)
        self.assertEqual(history_2.name, original_name)
        self.assertEqual(history_2.status, original_status)

        # 2nd changed field
        self.assertNotEqual(logs[0].user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(logs[0].project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(logs[0].category, {
            'id': str(self.category.id),
            'name': self.category.name})
        self.assertEqual(logs[0].field, None)
        self.assertEqual(logs[0].action, {
            'id': 'updated',
            'class': 'Category',
            'field': 'status',
            'value': self.category.status})
        history_1 = self.category.history.get(pk=logs[0].historical.get('id'))
        self.assertEqual(history_1.id, self.category.id)
        self.assertEqual(history_1.name, original_name)
        self.assertEqual(history_1.status, original_status)

        # History entry is only one per save
        self.assertEqual(history_1, history_2)



"""Base for superuser tools."""

from rest_framework.permissions import BasePermission


class IsSuperuser(BasePermission):
    """A permission to check if user is a superuser."""

    def has_permission(self, request, view):
        """
        Check if user is a superuser.

        Parameters
        ----------
        request : rest_framework.request.Request
            Object representing the request.
        view : rest_framework.views.APIView
            View that called the permission.

        Returns
        -------
        Boolean
            Indicating if user is a superuser.
        """
        return request.user and request.user.is_superuser



"""Tests for logger: model Subset."""

from django.test import TestCase

from geokey.core.models import LoggerHistory
from geokey.users.tests.model_factories import UserFactory
from geokey.projects.tests.model_factories import ProjectFactory
from geokey.subsets.tests.model_factories import SubsetFactory


class LogSubsetTest(TestCase):
    """Test model Subset."""

    def setUp(self):
        """Set up test."""
        self.user = UserFactory.create()
        self.project = ProjectFactory.create(**{
            'creator': self.user})
        self.subset = SubsetFactory.create(**{
            'creator': self.user,
            'project': self.project})

    def test_log_create(self):
        """Test when subset gets created."""
        log_count_init = LoggerHistory.objects.count()
        subset = SubsetFactory.create(**{
            'creator': self.user,
            'project': self.project})

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, None)
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, {
            'id': str(subset.id),
            'name': subset.name})
        self.assertEqual(log.action, {
            'id': 'created',
            'class': 'Subset'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_delete(self):
        """Test when subset gets deleted."""
        subset_id = self.subset.id
        subset_name = self.subset.name
        log_count_init = LoggerHistory.objects.count()
        self.subset.delete()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, None)
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, {
            'id': str(subset_id),
            'name': subset_name})
        self.assertEqual(log.action, {
            'id': 'deleted',
            'class': 'Subset'})
        self.assertEqual(log_count, log_count_init + 1)
        self.assertEqual(log.historical, None)

    def test_log_update_name(self):
        """Test when name changes."""
        log_count_init = LoggerHistory.objects.count()
        original_name = self.subset.name
        self.subset.name = '%s UPDATED' % self.subset.name
        self.subset.save()

        log = LoggerHistory.objects.last()
        log_count = LoggerHistory.objects.count()

        self.assertNotEqual(log.user, {
            'id': str(self.user.id),
            'display_name': self.user.display_name})
        self.assertEqual(log.project, {
            'id': str(self.project.id),
            'name': self.project.name})
        self.assertEqual(log.usergroup, None)
        self.assertEqual(log.category, None)
        self.assertEqual(log.field, None)
        self.assertEqual(log.location, None)
        self.assertEqual(log.observation, None)
        self.assertEqual(log.comment, None)
        self.assertEqual(log.subset, {
            'id': str(self.subset.id),
            'name': self.subset.name})
        self.assertEqual(log.action, {
            'id': 'updated',
            'class': 'Subset',
            'field': 'name'})
        self.assertEqual(log_count, log_count_init + 1)
        history = self.subset.history.get(pk=log.historical.get('id'))
        self.assertEqual(history.id, self.subset.id)
        self.assertEqual(history.name, original_name)



"""URLs for extensions."""

from os.path import dirname, isfile, join

from django.conf.urls import url, include

from geokey.extensions.base import extensions


urlpatterns = []

for extension in extensions:
    if isfile(join(dirname(__import__(extension).__file__), 'urls.py')):
        urls = '%s.urls' % extension
        urlpatterns.append(url(r'^', include(urls, namespace=extension)))



"""Views for subsets."""

from json import loads as json_loads

from django.core.urlresolvers import reverse
from django.views.generic import TemplateView
from django.utils.html import strip_tags
from django.utils.safestring import mark_safe
from django.shortcuts import redirect
from django.contrib import messages

from braces.views import LoginRequiredMixin

from geokey.core.decorators import handle_exceptions_for_admin
from geokey.projects.models import Project
from geokey.projects.views import ProjectContext

from .models import Subset


class SubsetList(LoginRequiredMixin, ProjectContext, TemplateView):

    """
    Displays the list of subsets in the project.
    """
    template_name = 'subsets/subset_list.html'


class SubsetCreate(LoginRequiredMixin, ProjectContext, TemplateView):

    """
    Provides the form to create a new subset.
    """
    template_name = 'subsets/subset_create.html'

    def post(self, request, project_id):
        """
        Creates the subset based on the data entered by the user.

        Parameters
        ----------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database

        Returns
        -------
        django.http.HttpResponseRedirect
            Redirects to subset create if subset is created, subset list if
            project is locked or it does not have any categories
        django.http.HttpResponse
            Rendered template, if project does not exist
        """

        data = request.POST
        context = self.get_context_data(project_id)
        project = context.get('project')

        if project:
            cannot_create = 'New subsets cannot be created.'

            if project.islocked:
                messages.error(
                    self.request,
                    'The project is locked. %s' % cannot_create
                )
                return redirect(
                    'admin:subset_create',
                    project_id=project_id
                )
            elif project.categories.count() == 0:
                messages.error(
                    self.request,
                    'The project has no categories. %s' % cannot_create
                )
                return redirect(
                    'admin:subset_create',
                    project_id=project_id
                )
            else:
                subset = Subset.objects.create(
                    name=strip_tags(data.get('name')),
                    description=strip_tags(data.get('description')),
                    creator=request.user,
                    project=project
                )

                add_another_url = reverse(
                    'admin:subset_create',
                    kwargs={
                        'project_id': project_id
                    }
                )

                messages.success(
                    self.request,
                    mark_safe('The subset has been created. <a href="%s">Add '
                              'another subset.</a>' % add_another_url)
                )

                return redirect(
                    'admin:subset_data',
                    project_id=project_id,
                    subset_id=subset.id
                )
        else:
            return self.render_to_response(context)


class SubsetContext(object):

    """
    Mixin that provides the context to render templates. The context contains
    a subset instance based on project_id and subset_id.
    """

    @handle_exceptions_for_admin
    def get_context_data(self, project_id, subset_id, *args, **kwargs):
        """
        Returns the context containing the project and subset instances.

        Parameters
        ----------
        project_id : int
            Identifies the project in the database
        subset_id : int
            Identifies the subset in the database

        Returns
        -------
        dict
            Context
        """

        project = Project.objects.as_admin(self.request.user, project_id)
        subset = project.subsets.get(pk=subset_id)

        return super(SubsetContext, self).get_context_data(
            project=project,
            subset=subset,
            *args,
            **kwargs
        )


class SubsetSettings(LoginRequiredMixin, SubsetContext, TemplateView):

    """
    Provides the form to update the subset settings.
    """
    template_name = 'subsets/subset_settings.html'

    def post(self, request, project_id, subset_id):
        """
        Updates the subset based on the data entered by the user.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database
        subset_id : int
            Identifies the subset in the database

        Returns
        -------
        django.http.HttpResponse
            Rendered template when subset updated
        django.http.HttpResponse
            Rendered template, if project or subset does not exist
        """

        data = request.POST
        context = self.get_context_data(project_id, subset_id)
        subset = context.get('subset')

        if subset:
            subset.name = strip_tags(data.get('name'))
            subset.description = strip_tags(data.get('description'))
            subset.save()

            messages.success(self.request, 'The subset has been updated.')

        return self.render_to_response(context)


class SubsetData(LoginRequiredMixin, SubsetContext, TemplateView):

    """
    Provides the form to change the filter settings for the subset.
    """
    template_name = 'subsets/subset_data.html'

    def post(self, request, project_id, subset_id):
        """
        Updates the subset filter based on the data entered by the user.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database
        subset_id : int
            Identifies the subset in the database

        Returns
        -------
        django.http.HttpResponse
            Rendered template when subset updated
        django.http.HttpResponse
            Rendered template, if project or subset does not exist
        """

        data = request.POST
        context = self.get_context_data(project_id, subset_id)
        subset = context.get('subset')

        if subset:
            cannot_modify = 'Subset data cannot be modified.'

            if subset.project.islocked:
                messages.error(
                    self.request,
                    'The project is locked. %s' % cannot_modify
                )
            elif subset.project.categories.count() == 0:
                messages.error(
                    self.request,
                    'The project has no categories. %s' % cannot_modify
                )
            else:
                if data['filters'] != '-1':
                    subset.filters = json_loads(data['filters'])
                    subset.save()

                    messages.success(
                        self.request,
                        'The subset has been updated.'
                    )

        return self.render_to_response(context)


class SubsetDelete(LoginRequiredMixin, SubsetContext, TemplateView):

    """
    Deletes the subset.
    """
    template_name = 'base.html'

    def get(self, request, project_id, subset_id):
        """
        Deletes the subset.

        Parameter
        ---------
        request : django.http.HttpRequest
            Object representing the request
        project_id : int
            Identifies the project in the database
        subset_id : int
            Identifies the subset in the database

        Returns
        -------
        django.http.HttpResponseRedirect
            Redirects to subset list if subset is deleted, subset settings if
            project is locked
        django.http.HttpResponse
            Rendered template, if project or subset does not exist
        """

        context = self.get_context_data(project_id, subset_id)
        subset = context.get('subset')

        if subset:
            if subset.project.islocked:
                messages.error(
                    self.request,
                    'The project is locked. Subset cannot be deleted.'
                )
                return redirect(
                    'admin:subset_settings',
                    project_id=project_id,
                    subset_id=subset_id
                )
            else:
                subset.delete()

                messages.success(self.request, 'The subset has been deleted.')
                return redirect('admin:subset_list', project_id=project_id)

        return self.render_to_response(context)



"""Mixins for superuser tools."""


class SuperuserMixin(object):
    """A mixin for superuser."""

    exception_message = 'No rights to access superuser tools.'

    def dispatch(self, request, *args, **kwargs):
        """
        Dispatch the request.

        Check if user is a superuser, display an error message if not.

        Parameters
        ----------
        request : django.http.HttpRequest
            Object representing the request.

        Returns
        -------
        django.http.HttpResponse
        """
        if not request.user.is_superuser:
            return self.render_to_response({
                'error': 'Permission denied.',
                'error_description': self.exception_message
            })

        return super(SuperuserMixin, self).dispatch(request, *args, **kwargs)
